#!/usr/bin/env py.test
# -*- coding: utf-8 -*-

from six.moves import xrange as range
import ufl
from ufl import *
#from ufl import product

from ffc.uflacs.backends.toy.toy_compiler import compile_expression

def compile_expression0(expr):
    code = ""
    return code

def compile_expression1(expr):
    code = compile_expression(expr, "")
    return code

def compile_tabulate_tensor_body(integral):
    # TODO: Handle measure type etc.
    if isinstance(integral, Form):
        integral, = integral.integrals()
    expr = integral.integrand()
    code = compile_expression(expr, "")
    return code

def test_interval_tabten_x_given(gtest):
    """Test code generation of body of the ufc function:

    void tabulate_tensor(
        double* A,
        const double * const * w,
        const double * coordinate_dofs,
        std::size_t num_points,
        const double * const * points
        ) const;

    """

    pre = """
    mock_cell mc;
    mc.fill_reference_interval(1);

    double * coordinate_dofs = mc.coordinate_dofs;
    double A[1] = { 0.0 };
    std::size_t num_points = 1;
    double points[1] = { 1.2 };

    // "Fetching" a point
    double * x = points;
    """

    post = """
    ASSERT_EQ(A[0], 1.2);
    """

    cell = interval

    x = SpatialCoordinate(cell)[0]
    expr = x
    integral = expr*dP

    code = compile_tabulate_tensor_body(integral)

    gtest.add(pre + code + post)

def test_interval_tabten_dg0_given(gtest):
    """Test code generation of body of the ufc function:

    void tabulate_tensor(
        double* A,
        const double * const * w,
        const double * coordinate_dofs,
        std::size_t num_points,
        const double * const * points
        ) const;

    with mock values for DG0/Real coefficients in w[][].

    """

    pre = """
    mock_cell mc;
    mc.fill_reference_interval(1);
    mc.coordinate_dofs[0*mc.geometric_dimension + 0] = 0.1;
    mc.coordinate_dofs[1*mc.geometric_dimension + 0] = 0.2;

    double A[1];
    memset(A, 0, sizeof(A));

    double w[2][2] = { { 1.2, 0.0 }, // second value here is unused
                       { 2.0, 3.0 } };

    double * coordinate_dofs = mc.coordinate_dofs;
    std::size_t num_points = 1;
    double points[1] = { 0.15 };

    // "Fetching" a point
    double * x = points;
    """

    post = """
    ASSERT_EQ(A[0], 0.15*1.2*(2.0+3.0));
    """

    cell = interval
    x = SpatialCoordinate(cell)[0]

    V = VectorElement("DG", cell, 0, dim=2)
    w0 = Constant(cell, count=0)
    w1 = Coefficient(V, count=1)

    expr = x*w0*(w1[0] + w1[1])

    integral = expr*dP

    code = compile_tabulate_tensor_body(integral)

    gtest.add(pre + code + post)

def xtest_interval_tabten_geometry_mappings(gtest):
    """Test code generation of body of the ufc function:

    void tabulate_tensor(
        double* A,
        const double * const * w,
        const double * coordinate_dofs
        ) const;
    """
    pass

def xtest_interval_geometry_expressions(gtest):
    """Test code generation of geometry expressions on an interval.

    """

    pre = """
    mock_cell mc;
    mc.fill_reference_interval(1);
    mc.coordinate_dofs[0*mc.geometric_dimension + 0] = 0.2;
    mc.coordinate_dofs[1*mc.geometric_dimension + 0] = 0.1;

    double * coordinate_dofs = mc.coordinate_dofs;
    double A[1];
    memset(A, 0, sizeof(A));

    """

    post = """
    // Check that geometric quantities are declared and computed correctly
    ASSERT_EQ(x[0], TODO);
    ASSERT_EQ(xi[0], TODO);
    ASSERT_EQ(J[0], -0.1);
    ASSERT_EQ(K[0], -10.0);
    ASSERT_EQ(detJ, -0.1);
    ASSERT_EQ(volume, 0.1);
    ASSERT_EQ(circumradius, TODO);

    // Check that geometric quantities have been placed in the output array
    std::size_t gd = mc.geometric_dimension;
    std::size_t td = mc.topological_dimension;
    double * AA = A;
    ASSERT_EQ(AA[0], x[0]); AA += gd;
    ASSERT_EQ(AA[0], xi[0]); AA += td;
    ASSERT_EQ(AA[0], J[0]); AA += gd * td;
    ASSERT_EQ(AA[0], K[0]); AA += td * gd;
    ASSERT_EQ(AA[0], detJ); AA += 1;
    ASSERT_EQ(AA[0], volume); AA += 1;
    ASSERT_EQ(AA[0], circumradius); AA += 1;
    """
    cell = interval

    gd = cell.geometric_dimension()
    td = cell.topological_dimension()

    values = []
    values.extend(SpatialCoordinate(cell)[i] for i in range(gd))
    values.extend(LocalCoordinate(cell)[i] for i in range(td))
    values.extend(GeometryJacobian(cell)[i, j] for i in range(gd) for i in range(td))
    values.extend(InverseGeometryJacobian(cell)[i, j] for i in range(td) for i in range(gd))
    values.append(GeometryJacobianDeterminant(cell))
    values.append(CellVolume(cell))
    values.append(Circumradius(cell))

    expr = as_vector(values)

    code = compile_expression1(expr)

    gtest.add(pre + code + post)

def test_tabulate_tensor_interval_facet(gtest):
    """Test code generation of body of the ufc function:

    void tabulate_tensor(
        double* A,
        const double * const * w,
        const double * coordinate_dofs,
        unsigned int facet) const;

    """

    pre = """
    mock_cell mc;
    mc.fill_reference_interval(1);
    mc.coordinate_dofs[0*mc.geometric_dimension + 0] = 0.1;
    mc.coordinate_dofs[1*mc.geometric_dimension + 0] = 0.2;

    double A[1];
    memset(A, 0, sizeof(A));

    double w[1][2] = { { 2.0, 3.0 } };

    double * coordinate_dofs = mc.coordinate_dofs;

    unsigned int facet = 0;
    """

    post = """
    // TODO
    """

    code = "// TODO"

    gtest.add(pre + code + post)

def test_tabulate_tensor_interval_interior_facet(gtest):
    """Test code generation of body of the ufc function:

    void tabulate_tensor(
        double* A,
        const double * const * w,
        const double * coordinate_dofs0,
        const double * coordinate_dofs1,
        unsigned int facet0,
        unsigned int facet1) const;
    """

    pre = """
    mock_cell mc0;
    mc0.fill_reference_interval(1);
    mc0.coordinate_dofs[0*mc0.geometric_dimension + 0] = 0.1;
    mc0.coordinate_dofs[1*mc0.geometric_dimension + 0] = 0.2;

    mock_cell mc1;
    mc1.fill_reference_interval(1);
    mc1.coordinate_dofs[0*mc1.geometric_dimension + 0] = 0.2;
    mc1.coordinate_dofs[1*mc1.geometric_dimension + 0] = 0.3;

    double A[2];
    memset(A, 0, sizeof(A));

    double w[1][4] = { { 2.0, 3.0,
                         4.0, 5.0 } };

    double * coordinate_dofs0 = mc0.coordinate_dofs;
    double * coordinate_dofs1 = mc1.coordinate_dofs;

    unsigned int facet0 = 1;
    unsigned int facet1 = 0;
    """

    post = """
    // TODO
    """

    code = "// TODO"

    gtest.add(pre + code + post)



# -*- coding: utf-8 -*-

# Copyright (C) 2009-2016 Marie E. Rognes
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Kristian B. Oelgaard 2010-2011
# Modified by Anders Logg 2013
# Modified by Lizao Li 2015, 2016

"""Code generation for evaluate_dof.

This module generates the functions evaluate_dof and evaluate_dofs.
These evaluate the degree of freedom (dof) number i and all degrees of
freedom for an element respectively.

Each dof L is assumed to act on a field f in the following manner:

  L(f) = w_{j, k} f_k(x_j)

where w is a set of weights, j is an index set corresponding to the
number of points involved in the evaluation of the functional, and k
is a multi-index set with rank corresponding to the value rank of the
function f.

For common degrees of freedom such as point evaluations and
directional component evaluations, there is just one point. However,
for various integral moments, the integrals are evaluated using
quadrature. The number of points therefore correspond to the
quadrature points.

The points x_j, weights w_{j, k} and components k are extracted from
FIAT (functional.pt_dict) in the intermediate representation stage.

"""

from collections import OrderedDict
import six

from ffc.cpp import format, remove_unused
from ffc.utils import pick_first
from ufl.permutation import build_component_numbering

__all__ = ["evaluate_dof_and_dofs", "affine_weights"]

# Prefetch formats:
comment = format["comment"]
declare = format["declaration"]
assign = format["assign"]
component = format["component"]
iadd = format["iadd"]
inner = format["inner product"]
add = format["addition"]
multiply = format["multiply"]
J = format["J"]
Jinv = format["inv(J)"]
detJ = format["det(J)"](None)
ret = format["return"]
f_i = format["argument dof num"]
f_values = format["argument values"]
f_double = format["float declaration"]
f_vals = format["dof vals"]
f_result = format["dof result"]
f_y = format["dof physical coordinates"]
f_x = format["coordinate_dofs"]
f_int = format["int declaration"]
f_X = format["dof X"]
f_D = format["dof D"]
f_W = format["dof W"]
f_copy = format["dof copy"]
f_r, f_s = format["free indices"][:2]
f_loop = format["generate loop"]

map_onto_physical = format["map onto physical"]


def evaluate_dof_and_dofs(ir):
    "Generate code for evaluate_dof and evaluate_dof."

    # Generate common code
    (reqs, cases) = _generate_common_code(ir)

    # Combine each case with returns for evaluate_dof and switch
    dof_cases = ["%s\n%s" % (c, ret(r)) for (c, r) in cases]
    dof_code = reqs + format["switch"](f_i, dof_cases, ret(format["float"](0.0)))

    # Construct dict with eval code as keys to remove duplicate eval code
    cases_opt = OrderedDict((case[0], []) for case in cases)
    for i, (evl, res) in enumerate(cases):
        cases_opt[evl].append((i, res))

    # Combine each case with assignments for evaluate_dofs
    dofs_code = reqs
    for evl, results in six.iteritems(cases_opt):
        dofs_code += evl + "\n"
        for i, res in results:
            dofs_code += format["assign"](component(f_values, i), res) + "\n"
    dofs_code = dofs_code.rstrip("\n")

    return (dof_code, dofs_code)


def _generate_common_code(ir):

    # Define necessary geometry information based on the ir
    reqs = _required_declarations(ir)

    # Extract variables
    mappings = ir["mappings"]
    offsets = ir["physical_offsets"]
    gdim = ir["geometric_dimension"]
    tdim = ir["topological_dimension"]

    # Generate bodies for each degree of freedom
    cases = [_generate_body(i, dof, mappings[i], gdim, tdim, offsets[i])
             for (i, dof) in enumerate(ir["dofs"])]

    return (reqs, cases)


def _required_declarations(ir):
    """Generate code for declaring required variables and geometry
    information.
    """
    # Enriched element, no dofs defined
    if not any(ir["dofs"]):
        return ""

    code = []
    gdim = ir["geometric_dimension"]
    tdim = ir["topological_dimension"]

    # Declare variable for storing the result and physical coordinates
    code.append(comment("Declare variables for result of evaluation"))
    code.append(declare(f_double, component(f_vals, ir["physical_value_size"])))
    code.append("")
    code.append(comment("Declare variable for physical coordinates"))
    code.append(declare(f_double, component(f_y, gdim)))
    code.append("")

    # Check whether Jacobians are necessary.
    needs_inverse_jacobian = any(["contravariant piola" in m
                                  for m in ir["mappings"]])
    needs_jacobian = any(["covariant piola" in m for m in ir["mappings"]])

    # Check if Jacobians are needed
    if not (needs_jacobian or needs_inverse_jacobian):
        return "\n".join(code)

    # Intermediate variable needed for multiple point dofs
    needs_temporary = any(len(dof) > 1 for dof in ir["dofs"])
    if needs_temporary:
        code.append(declare(f_double, f_result))

    # Add sufficient Jacobian information. Note: same criterion for
    # needing inverse Jacobian as for needing oriented Jacobian
    code.append(format["compute_jacobian"](tdim, gdim))
    if needs_inverse_jacobian:
        code.append("")
        code.append(format["compute_jacobian_inverse"](tdim, gdim))
        code.append("")
        code.append(format["orientation"](tdim, gdim))

    return "\n".join(code)


def _generate_body(i, dof, mapping, gdim, tdim, offset=0, result=f_result):
    "Generate code for a single dof."

    # EnrichedElement is handled by having [None, ..., None] dual basis
    if not dof:
        return (format["exception"]("evaluate_dof(s) for enriched element "
                                    "not implemented."), 0.0)

    points = list(dof.keys())

    # Generate different code if multiple points. (Otherwise ffc
    # compile time blows up.)
    if len(points) > 1:
        code = _generate_multiple_points_body(i, dof, mapping, gdim, tdim,
                                              offset, result)
        return (code, result)

    # Get weights for mapping reference point to physical
    x = points[0]
    w = affine_weights(tdim)(x)

    # Map point onto physical element: y = F_K(x)
    code = []
    for j in range(gdim):
        y = inner(w, [component(f_x(), (k * gdim + j,)) for k in range(tdim + 1)])
        code.append(assign(component(f_y, j), y))

    # Evaluate function at physical point
    code.append(format["evaluate function"])

    # Map function values to the reference element
    F = _change_variables(mapping, gdim, tdim, offset)

    # Simple affine functions deserve special case:
    if len(F) == 1:
        return ("\n".join(code), multiply([dof[x][0][0], F[0]]))

    # Flatten multiindices
    (index_map, _) = build_component_numbering([tdim] * len(dof[x][0][1]),
                                               ())

    # Take inner product between components and weights
    value = add([multiply([w, F[index_map[k]]]) for (w, k) in dof[x]])

    # Return eval code and value
    return ("\n".join(code), value)


def _generate_multiple_points_body(i, dof, mapping, gdim, tdim,
                                   offset=0, result=f_result):
    "Generate c++ for-loop for multiple points (integral bodies)"

    code = [assign(result, 0.0)]
    points = list(dof.keys())
    n = len(points)

    # Get number of tokens per point
    tokens = [dof[x] for x in points]
    len_tokens = pick_first([len(t) for t in tokens])

    # Declare points
    points = format["list"]([format["list"](x) for x in points])
    code += [declare(f_double, component(f_X(i), [n, tdim]),
                     points)]

    # Declare components
    components = [[c[0] for (w, c) in token] for token in tokens]
    components = format["list"]([format["list"](c) for c in components])
    code += [declare(f_int, component(f_D(i), [n, len_tokens]), components)]

    # Declare weights
    weights = [[w for (w, c) in token] for token in tokens]
    weights = format["list"]([format["list"](w) for w in weights])
    code += [declare(f_double, component(f_W(i), [n, len_tokens]), weights)]

    # Declare copy variable:
    code += [declare(f_double, component(f_copy(i), tdim))]

    # Add loop over points
    code += [comment("Loop over points")]

    # Map the points from the reference onto the physical element
    # assert(gdim == tdim), \
    #    "Integral moments not supported for manifolds (yet). Please fix"
    lines_r = [map_onto_physical[tdim][gdim] % {"i": i, "j": f_r}]

    # Evaluate function at physical point
    lines_r.append(comment("Evaluate function at physical point"))
    lines_r.append(format["evaluate function"])

    # Map function values to the reference element
    lines_r.append(comment("Map function to reference element"))
    F = _change_variables(mapping, gdim, tdim, offset)
    lines_r += [assign(component(f_copy(i), k), F_k)
                for (k, F_k) in enumerate(F)]

    # Add loop over directional components
    lines_r.append(comment("Loop over directions"))
    value = multiply([component(f_copy(i),
                                component(f_D(i), (f_r, f_s))),
                      component(f_W(i), (f_r, f_s))])
    # Add value from this point to total result
    lines_s = [iadd(result, value)]

    # Generate loop over s and add to r.
    loop_vars_s = [(f_s, 0, len_tokens)]
    lines_r += f_loop(lines_s, loop_vars_s)

    # Generate loop over r and add to code.
    loop_vars_r = [(f_r, 0, n)]
    code += f_loop(lines_r, loop_vars_r)

    code = "\n".join(code)
    return code


def _change_variables(mapping, gdim, tdim, offset):
    """Generate code for mapping function values according to
    'mapping' and offset.

    The basics of how to map a field from a physical to the reference
    domain. (For the inverse approach -- see interpolatevertexvalues)

    Let g be a field defined on a physical domain T with physical
    coordinates x. Let T_0 be a reference domain with coordinates
    X. Assume that F: T_0 -> T such that

      x = F(X)

    Let J be the Jacobian of F, i.e J = dx/dX and let K denote the
    inverse of the Jacobian K = J^{-1}. Then we (currently) have the
    following four types of mappings:

    'affine' mapping for g:

      G(X) = g(x)

    For vector fields g:

    'contravariant piola' mapping for g:

      G(X) = det(J) K g(x)   i.e  G_i(X) = det(J) K_ij g_j(x)

    'covariant piola' mapping for g:

      G(X) = J^T g(x)          i.e  G_i(X) = J^T_ij g(x) = J_ji g_j(x)

    'double covariant piola' mapping for g:

      G(X) = J^T g(x) J     i.e. G_il(X) = J_ji g_jk(x) J_kl

    'double contravariant piola' mapping for g:

      G(X) = det(J)^2 K g(x) K^T  i.e. G_il(X)=(detJ)^2 K_ij g_jk K_lk

    """

    # meg: Various mappings must be handled both here and in
    # interpolate_vertex_values. Could this be abstracted out?

    if mapping == "affine":
        return [component(f_vals, offset)]

    elif mapping == "contravariant piola":
        # Map each component from physical to reference using inverse
        # contravariant piola
        values = []
        for i in range(tdim):
            inv_jacobian_row = [Jinv(i, j, tdim, gdim) for j in range(gdim)]
            components = [component(f_vals, j + offset) for j in range(gdim)]
            values += [multiply([detJ, inner(inv_jacobian_row, components)])]
        return values

    elif mapping == "covariant piola":
        # Map each component from physical to reference using inverse
        # covariant piola
        values = []
        for i in range(tdim):
            jacobian_column = [J(j, i, gdim, tdim) for j in range(gdim)]
            components = [component(f_vals, j + offset) for j in range(gdim)]
            values += [inner(jacobian_column, components)]
        return values

    elif mapping == "double covariant piola":
        # physical to reference pullback as a covariant 2-tensor
        values = []
        for i in range(tdim):
            for l in range(tdim):
                values += [inner(
                    [inner([J(j, i, gdim, tdim) for j in range(gdim)],
                           [component(f_vals, j * tdim + k + offset)
                            for j in range(gdim)]) for k in range(gdim)],
                    [J(k, l, gdim, tdim) for k in range(gdim)])]
        return values

    elif mapping == "double contravariant piola":
        # physical to reference using double contravariant piola
        values = []
        for i in range(tdim):
            for l in range(tdim):
                values += [multiply([detJ, detJ, inner(
                    [inner([Jinv(i, j, gdim, tdim) for j in range(gdim)],
                           [component(f_vals, j * tdim + k + offset)
                            for j in range(gdim)]) for k in range(gdim)],
                    [Jinv(l, k, gdim, tdim) for k in range(gdim)])])]
        return values

    else:
        raise Exception("The mapping (%s) is not allowed" % mapping)


def affine_weights(dim):
    "Compute coefficents for mapping from reference to physical element"

    if dim == 1:
        return lambda x: (1.0 - x[0], x[0])
    elif dim == 2:
        return lambda x: (1.0 - x[0] - x[1], x[0], x[1])
    elif dim == 3:
        return lambda x: (1.0 - x[0] - x[1] - x[2], x[0], x[1], x[2])



# -*- coding: utf-8 -*-
"""
Tests of CNode formatting.
"""

from __future__ import print_function

from ffc.uflacs.language.cnodes import *


def test_cnode_expression_precedence():
    assert str(Add(1, 2)) == "1 + 2"
    assert str(Add(Add(1, 2), 3)) == "1 + 2 + 3"
    assert str(Add(1, Add(2, 3))) == "1 + (2 + 3)"
    assert str(Add(Add(1, 2), Add(3, 4))) == "1 + 2 + (3 + 4)"
    assert str(Add(Add(1, Add(2, 3)), 4)) == "1 + (2 + 3) + 4"

    assert str(Div(Mul(1, 2), 3)) == "1 * 2 / 3"
    assert str(Mul(Div(1, 2), 3)) == "1 / 2 * 3"
    assert str(Mul(1, Div(2, 3))) == "1 * (2 / 3)"
    assert str(Mul(1, Mul(2, Mul(3, 4)))) == "1 * (2 * (3 * 4))"
    assert str(Mul(Mul(Mul(1, 2), 3), 4)) == "1 * 2 * 3 * 4"
    assert str(Mul(Mul(1, Mul(2, 3)), 4)) == "1 * (2 * 3) * 4"

    assert str(Mul(Add(1, 2), Add(3, 4))) == "(1 + 2) * (3 + 4)"


def test_cnode_expressions():
    A = Symbol("A")
    B = Symbol("B")

    # Literals
    assert str(LiteralInt(123)) == "123"
    assert str(LiteralFloat(0.0)) == "0.0"
    assert str(LiteralFloat(1.0)) == "1.0"
    assert str(LiteralFloat(12.3)) == "12.3"  # 1.23e+01"

    # Variables
    # TODO: VariableAccess

    # Arrays
    assert str(ArrayAccess("A", (1,))) == "A[1]"
    assert str(ArrayAccess(A, (1, 2))) == "A[1][2]"
    assert str(A[1, 2, 3]) == "A[1][2][3]"
    assert str(ArrayAccess(ArrayDecl("double", "A", (2,)), 1)) == "A[1]"
    assert str(ArrayDecl("double", A, (2, 3))[1, 2]) == "A[1][2]"

    # FlattenedArray
    n = Symbol("n")
    decl = ArrayDecl("double", A, (4,))
    assert str(FlattenedArray(decl, strides=(2,), offset=3)[0]) == "A[3]"  # "A[3 + 2 * 0]"
    assert str(FlattenedArray(decl, strides=(2,))[0]) == "A[0]"  # "A[2 * 0]"
    decl = ArrayDecl("double", A, (2, 3, 4))
    flattened = FlattenedArray(decl, strides=(7, 8 * n, n - 1))
    #assert str(flattened[0, n, n * 7]) == "A[7 * 0 + 8 * n * n + (n - 1) * (n * 7)]"
    #assert str(flattened[0, n][n * 7]) == "A[7 * 0 + 8 * n * n + (n - 1) * (n * 7)]"
    #assert str(flattened[0][n][n * 7]) == "A[7 * 0 + 8 * n * n + (n - 1) * (n * 7)]"
    assert str(flattened[0, n, n * 7]) == "A[8 * n * n + (n - 1) * (n * 7)]"
    assert str(flattened[0, n][n * 7]) == "A[8 * n * n + (n - 1) * (n * 7)]"
    assert str(flattened[0][n][n * 7]) == "A[8 * n * n + (n - 1) * (n * 7)]"

    # Unary operators
    assert str(Pos(1)) == "+1"
    assert str(Neg(1)) == "-1"
    assert str(Not(1)) == "!1"
    assert str(BitNot(1)) == "~1"
    assert str(PreIncrement("i")) == "++i"
    assert str(PostIncrement("i")) == "i++"
    assert str(PreDecrement("i")) == "--i"
    assert str(PostDecrement("i")) == "i--"
    assert str(Call(Symbol("f"), [])) == "f()"
    assert str(Call("f", ["x", 3, Mul("y", 8.0)])) == "f(x, 3, y * 8.0)"
    assert str(Call("sin", Mul("theta", 3.0))) == "sin(theta * 3.0)"

    # Binary operators
    assert str(Add(1, 2)) == "1 + 2"
    assert str(Mul(1, 2)) == "1 * 2"
    assert str(Div(1, 2)) == "1 / 2"
    assert str(Sub(1, 2)) == "1 - 2"
    assert str(Mod(1, 2)) == "1 % 2"
    assert str(EQ(1, 2)) == "1 == 2"
    assert str(NE(1, 2)) == "1 != 2"
    assert str(LT(1, 2)) == "1 < 2"
    assert str(LE(1, 2)) == "1 <= 2"
    assert str(GT(1, 2)) == "1 > 2"
    assert str(GE(1, 2)) == "1 >= 2"
    assert str(And(1, 2)) == "1 && 2"
    assert str(Or(1, 2)) == "1 || 2"
    assert str(BitAnd(1, 2)) == "1 & 2"
    assert str(BitXor(1, 2)) == "1 ^ 2"
    assert str(BitOr(1, 2)) == "1 | 2"

    # Binary operators translated from python
    assert str(A + B) == "A + B"
    assert str(A * B) == "A * B"
    assert str(A / B) == "A / B"
    assert str(A - B) == "A - B"

    # Ternary operator
    assert str(Conditional(1, 2, 3)) == "1 ? 2 : 3"

    # N-ary "operators" simplify code generation
    assert str(Sum([1, 2, 3, 4])) == "1 + 2 + 3 + 4"
    assert str(Product([1, 2, 3, 4])) == "1 * 2 * 3 * 4"
    assert str(Product([Sum([1, 2, 3]), Sub(4, 5)])) == "(1 + 2 + 3) * (4 - 5)"

    # Custom expression
    assert str(Mul(VerbatimExpr("1 + std::foo(3)"), Add(5, 6))) == "(1 + std::foo(3)) * (5 + 6)"


def test_cnode_assignments():
    x = Symbol("x")
    y = Symbol("y")
    assert str(Assign(x, y)) == "x = y"
    assert str(AssignAdd(x, y)) == "x += y"
    assert str(AssignSub(x, y)) == "x -= y"
    assert str(AssignMul(x, y)) == "x *= y"
    assert str(AssignDiv(x, y)) == "x /= y"
    assert str(AssignMod(x, y)) == "x %= y"
    assert str(AssignLShift(x, y)) == "x <<= y"
    assert str(AssignRShift(x, y)) == "x >>= y"
    assert str(AssignAnd(x, y)) == "x &&= y"
    assert str(AssignOr(x, y)) == "x ||= y"
    assert str(AssignBitAnd(x, y)) == "x &= y"
    assert str(AssignBitXor(x, y)) == "x ^= y"
    assert str(AssignBitOr(x, y)) == "x |= y"


def test_cnode_variable_declarations():
    assert str(VariableDecl("foo", "x")) == "foo x;"
    assert str(VariableDecl("int", "n", 1)) == "int n = 1;"
    assert str(VariableDecl("double", "x", Mul("y", 3.0))) == "double x = y * 3.0;"


def test_1d_initializer_list():
    fmt = lambda v, s: format_indented_lines(build_initializer_lists(v, s, 0, str))
    assert fmt([], (0,)) == "{  }"
    assert fmt([1], (1,)) == "{ 1 }"
    assert fmt([1, 2], (2,)) == "{ 1, 2 }"
    assert fmt([1, 2, 3], (3,)) == "{ 1, 2, 3 }"


def test_nd_initializer_list_oneitem():
    fmt = lambda v, s: format_indented_lines(build_initializer_lists(v, s, 0, str))
    assert fmt([1], (1,)) == "{ 1 }"
    assert fmt([[1]], (1, 1)) == "{ { 1 } }"
    assert fmt([[[1]]], (1, 1, 1)) == "{ { { 1 } } }"
    assert fmt([[[[1]]]], (1, 1, 1, 1)) == "{ { { { 1 } } } }"


def test_nd_initializer_list_twoitems():
    fmt = lambda v, s: format_indented_lines(build_initializer_lists(v, s, 0, str))
    assert fmt([1, 2], (2,)) == "{ 1, 2 }"
    assert fmt([[1, 2]], (1, 2)) == "{ { 1, 2 } }"
    assert fmt([[[1, 2]]], (1, 1, 2)) == "{ { { 1, 2 } } }"
    assert fmt([[[[1, 2]]]], (1, 1, 1, 2)) == "{ { { { 1, 2 } } } }"
    # transpose it:
    assert fmt([[1], [2]], (2, 1)) == "{ { 1 },\n  { 2 } }"
    assert fmt([[[1], [2]]], (1, 2, 1)) == "{ { { 1 },\n    { 2 } } }"


def test_nd_initializer_list_twobytwoitems():
    fmt = lambda v, s: format_indented_lines(build_initializer_lists(v, s, 0, str))
    assert fmt([[1, 2], [3, 4]], (2, 2)) == "{ { 1, 2 },\n  { 3, 4 } }"
    assert fmt([[[1, 2], [3, 4]]], (1, 2, 2)) == "{ { { 1, 2 },\n    { 3, 4 } } }"
    assert fmt([[[[1, 2], [3, 4]]]], (1, 1, 2, 2)) == "{ { { { 1, 2 },\n      { 3, 4 } } } }"


def test_2d_initializer_list():
    assert format_indented_lines(build_initializer_lists([[1, 2, 3], [4, 5, 6]], (2, 3), 0, str)) == "{ { 1, 2, 3 },\n  { 4, 5, 6 } }"

    values = [[[1], [2]], [[3], [4]], [[5], [6]]]
    reference = """\
{ { { 1 },
    { 2 } },
  { { 3 },
    { 4 } },
  { { 5 },
    { 6 } } }"""
    assert format_indented_lines(build_initializer_lists(values, (3, 2, 1), 0, str)) == reference


def test_2d_numpy_initializer_list():
    import numpy
    values = [[1, 2, 3], [4, 5, 6]]
    array = numpy.asarray(values)
    sh = (2, 3)
    assert array.shape == sh
    fmt = lambda v, s: format_indented_lines(build_initializer_lists(v, s, 0, str))
    assert fmt(values, sh) == "{ { 1, 2, 3 },\n  { 4, 5, 6 } }"
    assert fmt(array, sh) == "{ { 1, 2, 3 },\n  { 4, 5, 6 } }"

    values = [[[1], [2]], [[3], [4]], [[5], [6]]]
    array = numpy.asarray(values)
    sh = (3, 2, 1)
    assert sh == array.shape
    reference = """\
{ { { 1 },
    { 2 } },
  { { 3 },
    { 4 } },
  { { 5 },
    { 6 } } }"""
    assert fmt(values, sh) == reference
    assert fmt(array, sh) == reference


def test_cnode_array_declarations():
    assert str(ArrayDecl("double", "x", 3)) == "double x[3];"
    assert str(ArrayDecl("double", "x", (3,))) == "double x[3];"
    assert str(ArrayDecl("double", "x", (3, 4))) == "double x[3][4];"

    assert str(ArrayDecl("double", "x", 3, [1., 2., 3.])) == "double x[3] = { 1.0, 2.0, 3.0 };"
    assert str(ArrayDecl("double", "x", (3,), [1., 2., 3.])) == "double x[3] = { 1.0, 2.0, 3.0 };"
    reference = """\
{
    double x[2][3] =
        { { 1.0, 2.0, 3.0 },
          { 4.0, 5.0, 6.0 } };
}"""
    assert str(Scope(ArrayDecl("double", "x", (2, 3), [[1., 2., 3.], [4., 5., 6.]]))) == reference


def test_cnode_comments():
    assert str(Comment("hello world")) == "// hello world"
    assert str(Comment("  hello\n world  ")) == "// hello\n// world"
    assert format_indented_lines(Indented(Comment("  hello\n world  ").cs_format())) == "    // hello\n    // world"


def test_cnode_statements():
    assert str(Break()) == "break;"
    assert str(Continue()) == "continue;"
    assert str(Return(Add(1, 2))) == "return 1 + 2;"
    assert str(Case("x")) == "case x:"
    assert str(Default()) == "default:"

    code = "for (std::vector<int>::iterator it = v.begin(); it != v.end(); ++it)\n{    /* foobar */\n}"
    assert str(VerbatimStatement(code)) == code


def test_cnode_loop_statements():
    body = [Assign("x", 3), AssignAdd("x", 5)]
    body_fmt = "{\n    x = 3;\n    x += 5;\n}"
    assert str(Scope(body)) == body_fmt
    assert str(Namespace("foo", body)) == "namespace foo\n" + body_fmt
    assert str(If(LT("x", 4.0), body)) == "if (x < 4.0)\n" + body_fmt
    assert str(ElseIf(LT("x", 4.0), body)) == "else if (x < 4.0)\n" + body_fmt
    assert str(Else(body)) == "else\n" + body_fmt
    assert str(While(LT("x", 4.0), body)) == "while (x < 4.0)\n" + body_fmt
    assert str(Do(LT("x", 4.0), body)) == "do\n" + body_fmt + " while (x < 4.0);"
    assert str(For(VariableDecl("int", "i", 0), LT("i", 4), PreIncrement("i"), body)) == "for (int i = 0; i < 4; ++i)\n" + body_fmt

    assert str(ForRange("i", 3, 7, Comment("body"))) == "for (int i = 3; i < 7; ++i)\n{\n    // body\n}"

    # Using assigns as both statements and expressions
    assert str(While(LT(AssignAdd("x", 4.0), 17.0), AssignAdd("A", "y"))) == "while ((x += 4.0) < 17.0)\n{\n    A += y;\n}"
    assert str(ForRange("i", 3, 7, AssignAdd("A", "i"))) == "for (int i = 3; i < 7; ++i)\n    A += i;"


def test_cnode_loop_helpers():
    i = Symbol("i")
    j = Symbol("j")
    A = Symbol("A")
    B = Symbol("B")
    C = Symbol("C")
    dst = A[i + 4 * j]
    src = 2.0 * B[j] * C[i]
    ranges = [(i, 0, 2), (j, 1, 3)]
    assert str(assign_loop(dst, src, ranges)) == """\
for (int i = 0; i < 2; ++i)
    for (int j = 1; j < 3; ++j)
        A[i + 4 * j] = 2.0 * B[j] * C[i];"""
    assert str(scale_loop(dst, src, ranges)) == """\
for (int i = 0; i < 2; ++i)
    for (int j = 1; j < 3; ++j)
        A[i + 4 * j] *= 2.0 * B[j] * C[i];"""
    assert str(accumulate_loop(dst, src, ranges)) == """\
for (int i = 0; i < 2; ++i)
    for (int j = 1; j < 3; ++j)
        A[i + 4 * j] += 2.0 * B[j] * C[i];"""


def test_cnode_switch_statements():
    assert str(Switch("x", [])) == "switch (x)\n{\n}"
    assert str(Switch("x", [], default=Assign("i", 3))) == "switch (x)\n{\ndefault:\n    {\n        i = 3;\n    }\n}"

    reference_switch = """switch (x)
{
case 1:
    {
        y = 3;
    }
    break;
case 2:
    {
        y = 4;
    }
    break;
default:
    {
        y = 5;
    }
}"""
    cnode_switch = str(Switch("x",
                              [(1, Assign("y", 3)), (2, Assign("y", 4)), ],
                              default=Assign("y", 5)))
    assert cnode_switch == reference_switch

    reference_switch = """switch (x)
{
case 1:
    y = 3;
case 2:
    y = 4;
default:
    y = 5;
}"""
    cnode_switch = str(Switch("x",
                              [(1, Assign("y", 3)), (2, Assign("y", 4)), ],
                              default=Assign("y", 5),
                              autobreak=False, autoscope=False))
    assert cnode_switch == reference_switch


def test_conceptual_tabulate_tensor():
    A = ArrayDecl("double", "A", (4, 6), values=0.0)
    code = StatementList([
        A,
        ForRange("q", 0, 2, [
            VariableDecl("double", "x", ArrayAccess("quadpoints", "q")),
            ForRange("i", 0, 4, [
                ForRange("j", 0, 6, [
                    AssignAdd(ArrayAccess(A, ("i", "j")),
                              Mul(ArrayAccess("FE0", ("q", "i")),
                                  ArrayAccess("FE1", ("q", "j"))))
                ])
            ])
        ])
    ])
    print(str(code))



# -*- coding: utf-8 -*-
# Copyright (C) 2011-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

"""Tools for computing various shapes of ufl expressions.

The total shape is the regular shape tuple plus the index shape tuple.
The index shape tuple is the tuple of index dimensions of the free indices
of the expression, sorted by the count of the free indices.

The total shape of a tensor valued expression ``A`` and
``A[*indices(len(A.ufl_shape))]`` is therefore the same.
"""


def compute_index_shape(v):
    """Compute the 'index shape' of v."""
    return v.ufl_index_dimensions


def compute_all_shapes(v):
    """Compute the tensor-, index-, and total shape of an expr.

    Returns (shape, size, index_shape, index_size, total_shape, total_size).
    """
    shape = v.ufl_shape
    index_shape = v.ufl_index_dimensions
    total_shape = shape + index_shape
    return (shape, index_shape, total_shape)


def total_shape(v):
    """Compute the total shape of an expr."""
    sh, ish, tsh = compute_all_shapes(v)
    return tsh



# -*- coding: utf-8 -*-
# Copyright (C) 2011-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

"""Linearized data structure for the computational graph."""

from ffc.uflacs.analysis.graph_vertices import build_graph_vertices
from ffc.uflacs.analysis.graph_symbols import build_graph_symbols


class Graph2(object):

    def __init__(self):

        self.nv = 0
        self.V = []
        self.e2i = {}

        self.expression_vertices = []

        self.V_shapes = []
        self.V_symbols = None  # Crs matrix
        self.total_unique_symbols = 0


def build_graph(expressions, DEBUG=False):

    # Make empty graph
    G = Graph2()

    # Populate with vertices
    G.e2i, G.V, G.expression_vertices = build_graph_vertices(expressions)
    G.nv = len(G.V)

    # Populate with symbols
    G.V_shapes, G.V_symbols, G.total_unique_symbols = \
        build_graph_symbols(G.V, G.e2i, DEBUG)

    if DEBUG:
        assert G.total_unique_symbols == len(set(G.V_symbols.data))

    return G



# -*- coding: utf-8 -*-


class DiffMarkerType:

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


DiffMissing = DiffMarkerType("<value missing>")
DiffEqual = DiffMarkerType("<equal>")

_default_recdiff_tolerance = 1e-5


def recdiff_dict(data1, data2, tolerance=_default_recdiff_tolerance):
    keys1 = set(data1.keys())
    keys2 = set(data2.keys())
    keys = keys1.intersection(keys2)
    diff = {}
    for k in keys1 - keys:
        diff[k] = (data1[k], DiffMissing)
    for k in keys2 - keys:
        diff[k] = (DiffMissing, data2[k])
    for k in keys:
        d1 = data1[k]
        d2 = data2[k]
        d = recdiff(d1, d2, tolerance)
        if d is not DiffEqual:
            diff[k] = d
    return diff or DiffEqual


def recdiff(data1, data2, tolerance=_default_recdiff_tolerance):
    if isinstance(data1, (float, int)) and isinstance(data2, (float, int)):
        # This approach allows numbers formatted as ints and floats interchangably as long as the values are equal
        delta = abs(data1 - data2)
        avg = (abs(data1) + abs(data2)) / 2.0

        if 0:
            # Using relative comparison, i.e. a tolerance of 1e-2 means one percent error is acceptable
            eps = tolerance * avg
            same = avg < 1e-14 or delta < eps
        else:
            # Using absolute comparison, this is what the old .out comparison does
            same = delta < tolerance

        return DiffEqual if same else (data1, data2)
    elif not isinstance(data1, type(data2)):
        return (data1, data2)
    elif isinstance(data1, dict):
        return recdiff_dict(data1, data2, tolerance)
    elif isinstance(data1, list):
        diff = [recdiff(d1, d2, tolerance) for (d1, d2) in zip(data1, data2)]
        return DiffEqual if all(d is DiffEqual for d in diff) else diff
    else:
        return DiffEqual if data1 == data2 else (data1, data2)


def _print(line):
    print(line)


def print_recdiff(diff, indent=0, printer=_print, prekey=""):

    if isinstance(diff, dict):
        for k in sorted(diff.keys()):
            key = str(k)
            if prekey:
                key = ".".join((prekey, key))
            printer("%s%s: " % ("  " * indent, key))
            print_recdiff(diff[k], indent + 1, printer, key)

    elif isinstance(diff, list):
        # Limiting this to lists of scalar values!
        for i, d in enumerate(diff):
            if isinstance(d, tuple):
                data1, data2 = d
                printer("%s%d: %s != %s" % ("  " * indent, i, data1, data2))

    elif isinstance(diff, tuple):
        assert len(diff) == 2
        data1, data2 = diff
        data1 = str(data1)
        data2 = str(data2)
        if len(data1) + len(data2) + 2 * indent + 4 > 70:
            printer("%s%s" % ("  " * indent, data1))
            printer("%s!=" % ("  " * indent))
            printer("%s%s" % ("  " * indent, data2))
        else:
            printer("%s%s != %s" % ("  " * indent, data1, data2))


# ---------- Unittest code
import unittest
# from recdiff import recdiff, print_recdiff, DiffEqual, DiffMissing


class RecDiffTestCase(unittest.TestCase):

    def assertEqual(self, a, b):
        if not (a == b):
            print(a)
            print(b)
            assert a == b

    def assertDiffEqual(self, diff):
        self.assertEqual(diff, DiffEqual)

    def test_recdiff_equal_items(self):
        self.assertDiffEqual(recdiff(1, 1))
        self.assertDiffEqual(recdiff(0, 0))
        self.assertDiffEqual(recdiff(0, 1e-15))
        self.assertDiffEqual(recdiff(1.1, 1.1 + 1e-7, tolerance=1e-6))
        self.assertDiffEqual(recdiff(1.1, 1.1 - 1e-7, tolerance=1e-6))
        self.assertDiffEqual(recdiff("foo", "foo"))

    def test_recdiff_not_equal_items(self):
        self.assertEqual(recdiff(1, 2), (1, 2))
        self.assertEqual(recdiff(0, 0.0001), (0, 0.0001))
        self.assertEqual(recdiff(0, 1e-13), (0, 1e-13))
        self.assertEqual(recdiff(1.1, 1.2 + 1e-7, tolerance=1e-6), (1.1, 1.2 + 1e-7))
        self.assertEqual(recdiff(1.1, 1.2 - 1e-7, tolerance=1e-6), (1.1, 1.2 - 1e-7))
        self.assertEqual(recdiff("foo", "bar"), ("foo", "bar"))

    def test_recdiff_equal_list(self):
        self.assertDiffEqual(recdiff([1, 2], [1, 2]))

    def test_recdiff_not_equal_list(self):
        self.assertEqual(recdiff([1, 2], [1, 3]), [DiffEqual, (2, 3)])

    def test_recdiff_equal_dict(self):
        self.assertDiffEqual(recdiff({1: 2}, {1: 2}))

    def test_recdiff_not_equal_dict(self):
        self.assertEqual(recdiff({1: 2, 2: 3}, {1: 3, 3: 4}), {1: (2, 3), 2: (3, DiffMissing), 3: (DiffMissing, 4)})

    def test_recdiff_equal_dict_hierarchy(self):
        self.assertDiffEqual(recdiff({1: {2: {3: 4, 5: 6}}}, {1: {2: {3: 4, 5: 6}}}))

    def test_recdiff_not_equal_dict_hierarchy(self):
        self.assertEqual(recdiff({1: {2: {3: 4, 5: 6}}}, {1: {2: {3: 4, 5: 7}}}), {1: {2: {5: (6, 7)}}})

    def test_example(self):
        form1 = {
            "num_coefficients": 2,
            "num_arguments": 2,
            "has_default_cell_integral": 1,
            "cell_integrals": {0: {"tabulate_tensor_input1": ["data"]}},
        }

        form2 = eval("""{
            "num_coefficients": 2,
            "rank": 2,
            "has_default_cell_integral": 0,
            "cell_integrals": { 0: { "tabulate_tensor_input1": ["data2"] } },
        }""")

        actual_diff = recdiff(form1, form2)
        if 0:
            print_recdiff(actual_diff)

        expected_diff = {
            #"num_coefficients": DiffEqual,
            "num_arguments": (2, DiffMissing),
            "rank": (DiffMissing, 2),
            "has_default_cell_integral": (1, 0),
            "cell_integrals": {0: {"tabulate_tensor_input1": [("data", "data2")]}},
        }
        self.assertEqual(actual_diff, expected_diff)


def main(a, b, tolerance=_default_recdiff_tolerance):
    print("Running diff on files %s and %s" % (a, b))
    a = eval(open(a).read())
    b = eval(open(b).read())
    d = recdiff(a, b, float(tolerance))
    print_recdiff(d)


if __name__ == "__main__":
    import sys
    args = sys.argv[1:]
    if not args:  # Hack to be able to use this as a script, TODO: do something nicer
        print("No arguments, running tests.")
        unittest.main()
    else:
        main(*args)



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2010-05-11
# Last changed: 2010-05-11

from __future__ import print_function

def print_table(values, title):
    "Print nicely formatted table."

    m = max([key[0] for key in values]) + 2
    n = max([key[1] for key in values]) + 2

    table = []
    for i in range(m):
        table.append(["" for j in range(n)])

    for i in range(m - 1):
        table[i + 1][0] = str(values[(i, 0)][0])

    for j in range(n - 1):
        table[0][j + 1] = str(values[(0, j)][1])

    for i in range(m - 1):
        for j in range(n - 1):
            value = values[(i, j)][2]
            if isinstance(value, float):
                value = "%.5g" % value
            table[i + 1][j + 1] = value

    table[0][0] = title

    column_sizes = [max([len(table[i][j]) for i in range(m)]) for j in range(n)]
    row_size = sum(column_sizes) + 3*(len(column_sizes) - 1) + 2

    print("")
    for i in range(m):
        print(" " + "-"*row_size)
        print("|", end="")
        for j in range(n):
            print(table[i][j] + " "*(column_sizes[j] - len(table[i][j])), end="")
            print("|", end="")
        print("")
    print(" " + "-"*row_size)
    print("")



# -*- coding: utf-8 -*-

# Copyright (C) 2007-2016 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Kristian B. Oelgaard 2010-2013
# Modified by Marie Rognes 2007-2012
# Modified by Peter Brune 2009
# Modified by Martin Sandve Alnæs, 2013
#
# First added:  2007-02-28
# Last changed: 2014-06-10

"Code snippets for code generation."

__all__ = ["comment_ufc", "comment_dolfin", "header_h", "header_c", "footer",
           "compute_jacobian", "compute_jacobian_inverse",
           "eval_basis_decl", "eval_basis_init", "eval_basis", "eval_basis_copy",
           "eval_derivs_decl", "eval_derivs_init", "eval_derivs", "eval_derivs_copy"]

__old__ = ["evaluate_f",
           "facet_determinant", "map_onto_physical",
           "fiat_coordinate_map", "transform_snippet",
           "scale_factor", "combinations_snippet",
           "normal_direction",
           "facet_normal", "ip_coordinates", "cell_volume", "circumradius",
           "facet_area", "min_facet_edge_length", "max_facet_edge_length",
           "orientation_snippet"]

__all__ += __old__

comment_ufc = """\
// This code conforms with the UFC specification version %(ufc_version)s
// and was automatically generated by FFC version %(ffc_version)s.
"""

comment_dolfin = """\
// This code conforms with the UFC specification version %(ufc_version)s
// and was automatically generated by FFC version %(ffc_version)s.
//
// This code was generated with the option '-l dolfin' and
// contains DOLFIN-specific wrappers that depend on DOLFIN.
"""

# Code snippets for headers and footers

header_h = """\
#ifndef __%(prefix_upper)s_H
#define __%(prefix_upper)s_H
"""

header_c = """\
#include "%(prefix)s.h"
"""

footer = """\
#endif
"""

# Code snippets for computing Jacobians

_compute_jacobian_interval_1d = """\
// Compute Jacobian
double J%(restriction)s[1];
compute_jacobian_interval_1d(J%(restriction)s, coordinate_dofs%(restriction)s);
"""

_compute_jacobian_interval_2d = """\
// Compute Jacobian
double J%(restriction)s[2];
compute_jacobian_interval_2d(J%(restriction)s, coordinate_dofs%(restriction)s);
"""

_compute_jacobian_interval_3d = """\
// Compute Jacobian
double J%(restriction)s[3];
compute_jacobian_interval_3d(J%(restriction)s, coordinate_dofs%(restriction)s);
"""

_compute_jacobian_triangle_2d = """\
// Compute Jacobian
double J%(restriction)s[4];
compute_jacobian_triangle_2d(J%(restriction)s, coordinate_dofs%(restriction)s);
"""

_compute_jacobian_triangle_3d = """\
// Compute Jacobian
double J%(restriction)s[6];
compute_jacobian_triangle_3d(J%(restriction)s, coordinate_dofs%(restriction)s);
"""

_compute_jacobian_tetrahedron_3d = """\
// Compute Jacobian
double J%(restriction)s[9];
compute_jacobian_tetrahedron_3d(J%(restriction)s, coordinate_dofs%(restriction)s);
"""

compute_jacobian = {1: {1: _compute_jacobian_interval_1d,
                        2: _compute_jacobian_interval_2d,
                        3: _compute_jacobian_interval_3d},
                    2: {2: _compute_jacobian_triangle_2d,
                        3: _compute_jacobian_triangle_3d},
                    3: {3: _compute_jacobian_tetrahedron_3d}}

# Code snippets for computing Jacobian inverses

_compute_jacobian_inverse_interval_1d = """\
// Compute Jacobian inverse and determinant
double K%(restriction)s[1];
double detJ%(restriction)s;
compute_jacobian_inverse_interval_1d(K%(restriction)s, detJ%(restriction)s, J%(restriction)s);
"""

_compute_jacobian_inverse_interval_2d = """\
// Compute Jacobian inverse and determinant
double K%(restriction)s[2];
double detJ%(restriction)s;
compute_jacobian_inverse_interval_2d(K%(restriction)s, detJ%(restriction)s, J%(restriction)s);
"""

_compute_jacobian_inverse_interval_3d = """\
// Compute Jacobian inverse and determinant
double K%(restriction)s[3];
double detJ%(restriction)s;
compute_jacobian_inverse_interval_3d(K%(restriction)s, detJ%(restriction)s, J%(restriction)s);
"""

_compute_jacobian_inverse_triangle_2d = """\
// Compute Jacobian inverse and determinant
double K%(restriction)s[4];
double detJ%(restriction)s;
compute_jacobian_inverse_triangle_2d(K%(restriction)s, detJ%(restriction)s, J%(restriction)s);
"""

_compute_jacobian_inverse_triangle_3d = """\
// Compute Jacobian inverse and determinant
double K%(restriction)s[6];
double detJ%(restriction)s;
compute_jacobian_inverse_triangle_3d(K%(restriction)s, detJ%(restriction)s, J%(restriction)s);
"""

_compute_jacobian_inverse_tetrahedron_3d = """\
// Compute Jacobian inverse and determinant
double K%(restriction)s[9];
double detJ%(restriction)s;
compute_jacobian_inverse_tetrahedron_3d(K%(restriction)s, detJ%(restriction)s, J%(restriction)s);
"""

compute_jacobian_inverse = {1: {1: _compute_jacobian_inverse_interval_1d,
                                2: _compute_jacobian_inverse_interval_2d,
                                3: _compute_jacobian_inverse_interval_3d},
                            2: {2: _compute_jacobian_inverse_triangle_2d,
                                3: _compute_jacobian_inverse_triangle_3d},
                            3: {3: _compute_jacobian_inverse_tetrahedron_3d}}

# Code snippet for scale factor
scale_factor = """\
// Set scale factor
const double det = std::abs(detJ);"""

# FIXME: Old stuff below that should be cleaned up or moved to ufc_geometry.h

orientation_snippet = """
// Check orientation
if (cell_orientation%(restriction)s == -1)
  throw std::runtime_error("cell orientation must be defined (not -1)");
// (If cell_orientation == 1 = down, multiply det(J) by -1)
else if (cell_orientation%(restriction)s == 1)
  detJ%(restriction)s *= -1;
"""

evaluate_f = "f.evaluate(vals, y, c);"

_facet_determinant_1D = """\
// Facet determinant 1D (vertex)
const double det = 1.0;"""

_facet_determinant_2D = """\
// Get vertices on edge
static unsigned int edge_vertices[3][2] = {{1, 2}, {0, 2}, {0, 1}};
const unsigned int v0 = edge_vertices[facet%(restriction)s][0];
const unsigned int v1 = edge_vertices[facet%(restriction)s][1];

// Compute scale factor (length of edge scaled by length of reference interval)
const double dx0 = coordinate_dofs%(restriction)s[2*v1 + 0] - coordinate_dofs%(restriction)s[2*v0 + 0];
const double dx1 = coordinate_dofs%(restriction)s[2*v1 + 1] - coordinate_dofs%(restriction)s[2*v0 + 1];
const double det = std::sqrt(dx0*dx0 + dx1*dx1);
"""

_facet_determinant_2D_1D = """\
// Facet determinant 1D in 2D (vertex)
const double det = 1.0;
"""

_facet_determinant_3D = """\
// Get vertices on face
static unsigned int face_vertices[4][3] = {{1, 2, 3}, {0, 2, 3}, {0, 1, 3}, {0, 1, 2}};
const unsigned int v0 = face_vertices[facet%(restriction)s][0];
const unsigned int v1 = face_vertices[facet%(restriction)s][1];
const unsigned int v2 = face_vertices[facet%(restriction)s][2];

// Compute scale factor (area of face scaled by area of reference triangle)
const double a0 = (coordinate_dofs%(restriction)s[3*v0 + 1]*coordinate_dofs%(restriction)s[3*v1 + 2]  + coordinate_dofs%(restriction)s[3*v0 + 2]*coordinate_dofs%(restriction)s[3*v2 + 1]  + coordinate_dofs%(restriction)s[3*v1 + 1]*coordinate_dofs%(restriction)s[3*v2 + 2]) - (coordinate_dofs%(restriction)s[3*v2 + 1]*coordinate_dofs%(restriction)s[3*v1 + 2] + coordinate_dofs%(restriction)s[3*v2 + 2]*coordinate_dofs%(restriction)s[3*v0 + 1] + coordinate_dofs%(restriction)s[3*v1 + 1]*coordinate_dofs%(restriction)s[3*v0 + 2]);

const double a1 = (coordinate_dofs%(restriction)s[3*v0 + 2]*coordinate_dofs%(restriction)s[3*v1 + 0]  + coordinate_dofs%(restriction)s[3*v0 + 0]*coordinate_dofs%(restriction)s[3*v2 + 2] + coordinate_dofs%(restriction)s[3*v1 + 2]*coordinate_dofs%(restriction)s[3*v2 + 0]) - (coordinate_dofs%(restriction)s[3*v2 + 2]*coordinate_dofs%(restriction)s[3*v1 + 0]  + coordinate_dofs%(restriction)s[3*v2 + 0]*coordinate_dofs%(restriction)s[3*v0 + 2] + coordinate_dofs%(restriction)s[3*v1 + 2]*coordinate_dofs%(restriction)s[3*v0 + 0]);

const double a2 = (coordinate_dofs%(restriction)s[3*v0 + 0]*coordinate_dofs%(restriction)s[3*v1 + 1]  + coordinate_dofs%(restriction)s[3*v0 + 1]*coordinate_dofs%(restriction)s[3*v2 + 0]  + coordinate_dofs%(restriction)s[3*v1 + 0]*coordinate_dofs%(restriction)s[3*v2 + 1]) - (coordinate_dofs%(restriction)s[3*v2 + 0]*coordinate_dofs%(restriction)s[3*v1 + 1]  + coordinate_dofs%(restriction)s[3*v2 + 1]*coordinate_dofs%(restriction)s[3*v0 + 0]  + coordinate_dofs%(restriction)s[3*v1 + 0]*coordinate_dofs%(restriction)s[3*v0 + 1]);

const double det = std::sqrt(a0*a0 + a1*a1 + a2*a2);
"""

_facet_determinant_3D_2D = """\
// Facet determinant 2D in 3D (edge)
// Get vertices on edge
static unsigned int edge_vertices[3][2] = {{1, 2}, {0, 2}, {0, 1}};
const unsigned int v0 = edge_vertices[facet%(restriction)s][0];
const unsigned int v1 = edge_vertices[facet%(restriction)s][1];

// Compute scale factor (length of edge scaled by length of reference interval)
const double dx0 = coordinate_dofs%(restriction)s[3*v1 + 0] - coordinate_dofs%(restriction)s[3*v0 + 0];
const double dx1 = coordinate_dofs%(restriction)s[3*v1 + 1] - coordinate_dofs%(restriction)s[3*v0 + 1];
const double dx2 = coordinate_dofs%(restriction)s[3*v1 + 2] - coordinate_dofs%(restriction)s[3*v0 + 2];
const double det = std::sqrt(dx0*dx0 + dx1*dx1 + dx2*dx2);
"""

_facet_determinant_3D_1D = """\
// Facet determinant 1D in 3D (vertex)
const double det = 1.0;
"""

_normal_direction_1D = """\
const bool direction = facet%(restriction)s == 0 ? coordinate_dofs%(restriction)s[0] > coordinate_dofs%(restriction)s[1] : coordinate_dofs%(restriction)s[1] > coordinate_dofs%(restriction)s[0];
"""

_normal_direction_2D = """\
const bool direction = dx1*(coordinate_dofs%(restriction)s[2*%(facet)s] - coordinate_dofs%(restriction)s[2*v0]) - dx0*(coordinate_dofs%(restriction)s[2*%(facet)s + 1] - coordinate_dofs%(restriction)s[2*v0 + 1]) < 0;
"""

_normal_direction_3D = """\
const bool direction = a0*(coordinate_dofs%(restriction)s[3*%(facet)s] - coordinate_dofs%(restriction)s[3*v0]) + a1*(coordinate_dofs%(restriction)s[3*%(facet)s + 1] - coordinate_dofs%(restriction)s[3*v0 + 1])  + a2*(coordinate_dofs%(restriction)s[3*%(facet)s + 2] - coordinate_dofs%(restriction)s[3*v0 + 2]) < 0;
"""

# MER: Coding all up in _facet_normal_ND_M_D for now; these are
# therefore empty.
_normal_direction_2D_1D = ""
_normal_direction_3D_2D = ""
_normal_direction_3D_1D = ""

_facet_normal_1D = """
// Facet normals are 1.0 or -1.0:   (-1.0) <-- X------X --> (1.0)
const double n%(restriction)s = %(direction)sdirection ? 1.0 : -1.0;"""

_facet_normal_2D = """\
// Compute facet normals from the facet scale factor constants
const double n%(restriction)s0 = %(direction)sdirection ? dx1 / det : -dx1 / det;
const double n%(restriction)s1 = %(direction)sdirection ? -dx0 / det : dx0 / det;"""

_facet_normal_2D_1D = """
// Compute facet normal
double n%(restriction)s0 = 0.0;
double n%(restriction)s1 = 0.0;
if (facet%(restriction)s == 0)
{
  n%(restriction)s0 = coordinate_dofs%(restriction)s[0] - coordinate_dofs%(restriction)s[2];
  n%(restriction)s1 = coordinate_dofs%(restriction)s[1] - coordinate_dofs%(restriction)s[3];
}
else
{
  n%(restriction)s0 = coordinate_dofs%(restriction)s[2] - coordinate_dofs%(restriction)s[0];
  n%(restriction)s1 = coordinate_dofs%(restriction)s[3] - coordinate_dofs%(restriction)s[1];
}
const double n%(restriction)s_length = std::sqrt(n%(restriction)s0*n%(restriction)s0 + n%(restriction)s1*n%(restriction)s1);
n%(restriction)s0 /= n%(restriction)s_length;
n%(restriction)s1 /= n%(restriction)s_length;
"""

_facet_normal_3D = """
const double n%(restriction)s0 = %(direction)sdirection ? a0 / det : -a0 / det;
const double n%(restriction)s1 = %(direction)sdirection ? a1 / det : -a1 / det;
const double n%(restriction)s2 = %(direction)sdirection ? a2 / det : -a2 / det;"""

_facet_normal_3D_2D = """
// Compute facet normal for triangles in 3D
const unsigned int vertex%(restriction)s0 = facet%(restriction)s;

// Get coordinates corresponding the vertex opposite this
// static unsigned int edge_vertices[3][2] = {{1, 2}, {0, 2}, {0, 1}};
const unsigned int vertex%(restriction)s1 = edge_vertices[facet%(restriction)s][0];
const unsigned int vertex%(restriction)s2 = edge_vertices[facet%(restriction)s][1];

// Define vectors n = (p2 - p0) and t = normalized (p2 - p1)
double n%(restriction)s0 = coordinate_dofs%(restriction)s[3*vertex%(restriction)s2 + 0] - coordinate_dofs%(restriction)s[3*vertex%(restriction)s0 + 0];
double n%(restriction)s1 = coordinate_dofs%(restriction)s[3*vertex%(restriction)s2 + 1] - coordinate_dofs%(restriction)s[3*vertex%(restriction)s0 + 1];
double n%(restriction)s2 = coordinate_dofs%(restriction)s[3*vertex%(restriction)s2 + 2] - coordinate_dofs%(restriction)s[3*vertex%(restriction)s0 + 2];

double t%(restriction)s0 = coordinate_dofs%(restriction)s[3*vertex%(restriction)s2 + 0] - coordinate_dofs%(restriction)s[3*vertex%(restriction)s1 + 0];
double t%(restriction)s1 = coordinate_dofs%(restriction)s[3*vertex%(restriction)s2 + 1] - coordinate_dofs%(restriction)s[3*vertex%(restriction)s1 + 1];
double t%(restriction)s2 = coordinate_dofs%(restriction)s[3*vertex%(restriction)s2 + 2] - coordinate_dofs%(restriction)s[3*vertex%(restriction)s1 + 2];
const double t%(restriction)s_length = std::sqrt(t%(restriction)s0*t%(restriction)s0 + t%(restriction)s1*t%(restriction)s1 + t%(restriction)s2*t%(restriction)s2);
t%(restriction)s0 /= t%(restriction)s_length;
t%(restriction)s1 /= t%(restriction)s_length;
t%(restriction)s2 /= t%(restriction)s_length;

// Subtract, the projection of (p2  - p0) onto (p2 - p1), from (p2 - p0)
const double ndott%(restriction)s = t%(restriction)s0*n%(restriction)s0 + t%(restriction)s1*n%(restriction)s1 + t%(restriction)s2*n%(restriction)s2;
n%(restriction)s0 -= ndott%(restriction)s*t%(restriction)s0;
n%(restriction)s1 -= ndott%(restriction)s*t%(restriction)s1;
n%(restriction)s2 -= ndott%(restriction)s*t%(restriction)s2;
const double n%(restriction)s_length = std::sqrt(n%(restriction)s0*n%(restriction)s0 + n%(restriction)s1*n%(restriction)s1 + n%(restriction)s2*n%(restriction)s2);

// Normalize
n%(restriction)s0 /= n%(restriction)s_length;
n%(restriction)s1 /= n%(restriction)s_length;
n%(restriction)s2 /= n%(restriction)s_length;
"""

_facet_normal_3D_1D = """
// Compute facet normal
double n%(restriction)s0 = 0.0;
double n%(restriction)s1 = 0.0;
double n%(restriction)s2 = 0.0;
if (facet%(restriction)s == 0)
{
  n%(restriction)s0 = coordinate_dofs%(restriction)s[0] - coordinate_dofs%(restriction)s[3];
  n%(restriction)s1 = coordinate_dofs%(restriction)s[1] - coordinate_dofs%(restriction)s[4];
  n%(restriction)s1 = coordinate_dofs%(restriction)s[2] - coordinate_dofs%(restriction)s[5];
}
else
{
  n%(restriction)s0 = coordinate_dofs%(restriction)s[3] - coordinate_dofs%(restriction)s[0];
  n%(restriction)s1 = coordinate_dofs%(restriction)s[4] - coordinate_dofs%(restriction)s[1];
  n%(restriction)s1 = coordinate_dofs%(restriction)s[5] - coordinate_dofs%(restriction)s[2];
}
const double n%(restriction)s_length = std::sqrt(n%(restriction)s0*n%(restriction)s0 + n%(restriction)s1*n%(restriction)s1 + n%(restriction)s2*n%(restriction)s2);
n%(restriction)s0 /= n%(restriction)s_length;
n%(restriction)s1 /= n%(restriction)s_length;
n%(restriction)s2 /= n%(restriction)s_length;
"""

_cell_volume_1D = """\
// Compute cell volume
const double volume%(restriction)s = std::abs(detJ%(restriction)s);
"""

_cell_volume_2D = """\
// Compute cell volume
const double volume%(restriction)s = std::abs(detJ%(restriction)s)/2.0;
"""

_cell_volume_2D_1D = """\
// Compute cell volume of interval in 2D
const double volume%(restriction)s = std::abs(detJ%(restriction)s);
"""

_cell_volume_3D = """\
// Compute cell volume
const double volume%(restriction)s = std::abs(detJ%(restriction)s)/6.0;
"""

_cell_volume_3D_1D = """\
// Compute cell volume of interval in 3D
const double volume%(restriction)s = std::abs(detJ%(restriction)s);
"""

_cell_volume_3D_2D = """\
// Compute cell volume of triangle in 3D
const double volume%(restriction)s = std::abs(detJ%(restriction)s)/2.0;
"""

_circumradius_1D = """\
// Compute circumradius; in 1D it is equal to half the cell length
const double circumradius%(restriction)s = std::abs(detJ%(restriction)s)/2.0;
"""

_circumradius_2D = """\
// Compute circumradius of triangle in 2D
const double v1v2%(restriction)s  = std::sqrt((coordinate_dofs%(restriction)s[4] - coordinate_dofs%(restriction)s[2])*(coordinate_dofs%(restriction)s[4] - coordinate_dofs%(restriction)s[2]) + (coordinate_dofs%(restriction)s[5] - coordinate_dofs%(restriction)s[3])*(coordinate_dofs%(restriction)s[5] - coordinate_dofs%(restriction)s[3]) );
const double v0v2%(restriction)s  = std::sqrt(J%(restriction)s[3]*J%(restriction)s[3] + J%(restriction)s[1]*J%(restriction)s[1]);
const double v0v1%(restriction)s  = std::sqrt(J%(restriction)s[0]*J%(restriction)s[0] + J%(restriction)s[2]*J%(restriction)s[2]);
const double circumradius%(restriction)s = 0.25*(v1v2%(restriction)s*v0v2%(restriction)s*v0v1%(restriction)s)/(volume%(restriction)s);
"""

_circumradius_2D_1D = """\
// Compute circumradius of interval in 3D (1/2 volume)
const double circumradius%(restriction)s = std::abs(detJ%(restriction)s)/2.0;
"""

_circumradius_3D = """\
// Compute circumradius
const double v1v2%(restriction)s  = std::sqrt( (coordinate_dofs%(restriction)s[6] - coordinate_dofs%(restriction)s[3])*(coordinate_dofs%(restriction)s[6] - coordinate_dofs%(restriction)s[3]) + (coordinate_dofs%(restriction)s[7] - coordinate_dofs%(restriction)s[4])*(coordinate_dofs%(restriction)s[7] - coordinate_dofs%(restriction)s[4]) + (coordinate_dofs%(restriction)s[8] - coordinate_dofs%(restriction)s[5])*(coordinate_dofs%(restriction)s[8] - coordinate_dofs%(restriction)s[5]) );
const double v0v2%(restriction)s  = std::sqrt(J%(restriction)s[1]*J%(restriction)s[1] + J%(restriction)s[4]*J%(restriction)s[4] + J%(restriction)s[7]*J%(restriction)s[7]);
const double v0v1%(restriction)s  = std::sqrt(J%(restriction)s[0]*J%(restriction)s[0] + J%(restriction)s[3]*J%(restriction)s[3] + J%(restriction)s[6]*J%(restriction)s[6]);
const double v0v3%(restriction)s  = std::sqrt(J%(restriction)s[2]*J%(restriction)s[2] + J%(restriction)s[5]*J%(restriction)s[5] + J%(restriction)s[8]*J%(restriction)s[8]);
const double v1v3%(restriction)s  = std::sqrt( (coordinate_dofs%(restriction)s[9] - coordinate_dofs%(restriction)s[3])*(coordinate_dofs%(restriction)s[9] - coordinate_dofs%(restriction)s[3]) + (coordinate_dofs%(restriction)s[10] - coordinate_dofs%(restriction)s[4])*(coordinate_dofs%(restriction)s[10] - coordinate_dofs%(restriction)s[4]) + (coordinate_dofs%(restriction)s[11] - coordinate_dofs%(restriction)s[5])*(coordinate_dofs%(restriction)s[11] - coordinate_dofs%(restriction)s[5]) );
const double v2v3%(restriction)s  = std::sqrt( (coordinate_dofs%(restriction)s[9] - coordinate_dofs%(restriction)s[6])*(coordinate_dofs%(restriction)s[9] - coordinate_dofs%(restriction)s[6]) + (coordinate_dofs%(restriction)s[10] - coordinate_dofs%(restriction)s[7])*(coordinate_dofs%(restriction)s[10] - coordinate_dofs%(restriction)s[7]) + (coordinate_dofs%(restriction)s[11] - coordinate_dofs%(restriction)s[8])*(coordinate_dofs%(restriction)s[11] - coordinate_dofs%(restriction)s[8]) );
const  double la%(restriction)s   = v1v2%(restriction)s*v0v3%(restriction)s;
const  double lb%(restriction)s   = v0v2%(restriction)s*v1v3%(restriction)s;
const  double lc%(restriction)s   = v0v1%(restriction)s*v2v3%(restriction)s;
const  double s%(restriction)s    = 0.5*(la%(restriction)s+lb%(restriction)s+lc%(restriction)s);
const  double area%(restriction)s = std::sqrt(s%(restriction)s*(s%(restriction)s-la%(restriction)s)*(s%(restriction)s-lb%(restriction)s)*(s%(restriction)s-lc%(restriction)s));
const double circumradius%(restriction)s = area%(restriction)s / ( 6.0*volume%(restriction)s );
"""

_circumradius_3D_1D = """\
// Compute circumradius of interval in 3D (1/2 volume)
const double circumradius%(restriction)s = std::abs(detJ%(restriction)s)/2.0;
"""

_circumradius_3D_2D = """\
// Compute circumradius of triangle in 3D
const double v1v2%(restriction)s  = std::sqrt( (coordinate_dofs%(restriction)s[6] - coordinate_dofs%(restriction)s[3])*(coordinate_dofs%(restriction)s[6] - coordinate_dofs%(restriction)s[3]) + (coordinate_dofs%(restriction)s[7] - coordinate_dofs%(restriction)s[4])*(coordinate_dofs%(restriction)s[7] - coordinate_dofs%(restriction)s[4]) + (coordinate_dofs%(restriction)s[8] - coordinate_dofs%(restriction)s[5])*(coordinate_dofs%(restriction)s[8] - coordinate_dofs%(restriction)s[5]));
const double v0v2%(restriction)s = std::sqrt( J%(restriction)s[3]*J%(restriction)s[3] + J%(restriction)s[1]*J%(restriction)s[1] + J%(restriction)s[5]*J%(restriction)s[5]);
const double v0v1%(restriction)s = std::sqrt( J%(restriction)s[0]*J%(restriction)s[0] + J%(restriction)s[2]*J%(restriction)s[2] + J%(restriction)s[4]*J%(restriction)s[4]);
const double circumradius%(restriction)s = 0.25*(v1v2%(restriction)s*v0v2%(restriction)s*v0v1%(restriction)s)/(volume%(restriction)s);
"""

_facet_area_1D = """\
// Facet area (FIXME: Should this be 0.0?)
const double facet_area = 1.0;"""

_facet_area_2D = """\
// Facet area
const double facet_area = det;"""

_facet_area_2D_1D = """\
// Facet area
const double facet_area = 1.0;"""

_facet_area_3D = """\
// Facet area (divide by two because 'det' is scaled by area of reference triangle)
const double facet_area = det/2.0;"""

_facet_area_3D_1D = """\
// Facet area
const double facet_area = 1.0;"""

_facet_area_3D_2D = """\
// Facet area
const double facet_area = det;"""

evaluate_basis_dofmap = """\
unsigned int element = 0;
unsigned int tmp = 0;
for (unsigned int j = 0; j < %d; j++)
{
  if (tmp +  dofs_per_element[j] > i)
  {
    i -= tmp;
    element = element_types[j];
    break;
  }
  else
    tmp += dofs_per_element[j];
}"""

_min_facet_edge_length_3D = """\
// Min edge length of facet
double min_facet_edge_length;
compute_min_facet_edge_length_tetrahedron_3d(min_facet_edge_length, facet%(restriction)s, coordinate_dofs%(restriction)s);
"""

_max_facet_edge_length_3D = """\
// Max edge length of facet
double max_facet_edge_length;
compute_max_facet_edge_length_tetrahedron_3d(max_facet_edge_length, facet%(restriction)s, coordinate_dofs%(restriction)s);
"""

# FIXME: This is dead slow because of all the new calls
# Used in evaluate_basis_derivatives. For second order derivatives in 2D it will
# generate the combinations: [(0, 0), (0, 1), (1, 0), (1, 1)] (i.e., xx, xy, yx, yy)
# which will also be the ordering of derivatives in the return value.
combinations_snippet = """\
// Declare two dimensional array that holds combinations of derivatives and initialise
unsigned int %(combinations)s[%(max_num_derivatives)s][%(max_degree)s];
for (unsigned int row = 0; row < %(max_num_derivatives)s; row++)
{
  for (unsigned int col = 0; col < %(max_degree)s; col++)
    %(combinations)s[row][col] = 0;
}

// Generate combinations of derivatives
for (unsigned int row = 1; row < %(num_derivatives)s; row++)
{
  for (unsigned int num = 0; num < row; num++)
  {
    for (unsigned int col = %(n)s-1; col+1 > 0; col--)
    {
      if (%(combinations)s[row][col] + 1 > %(dimension-1)s)
        %(combinations)s[row][col] = 0;
      else
      {
        %(combinations)s[row][col] += 1;
        break;
      }
    }
  }
}"""


def _transform_snippet(tdim, gdim):

    if tdim == gdim:
        _t = ""
        _g = ""
    else:
        _t = "_t"
        _g = "_g"

    # Matricize K_ij -> {K_ij}
    matrix = "{{" + "}, {".join([", ".join(["K[%d]" % (t * gdim + g)
                                            for g in range(gdim)])
                                 for t in range(tdim)]) + "}};\n\n"
    snippet = """\
// Compute inverse of Jacobian
const double %%(K)s[%d][%d] = %s""" % (tdim, gdim, matrix)

    snippet += """// Declare transformation matrix
// Declare pointer to two dimensional array and initialise
double %%(transform)s[%%(max_g_deriv)s][%%(max_t_deriv)s];
for (unsigned int j = 0; j < %%(num_derivatives)s%(g)s; j++)
{
  for (unsigned int k = 0; k < %%(num_derivatives)s%(t)s; k++)
    %%(transform)s[j][k] = 1;
}

// Construct transformation matrix
for (unsigned int row = 0; row < %%(num_derivatives)s%(g)s; row++)
{
  for (unsigned int col = 0; col < %%(num_derivatives)s%(t)s; col++)
  {
    for (unsigned int k = 0; k < %%(n)s; k++)
      %%(transform)s[row][col] *= %%(K)s[%%(combinations)s%(t)s[col][k]][%%(combinations)s%(g)s[row][k]];
  }
}""" % {"t": _t, "g": _g}

    return snippet


# Codesnippets used in evaluate_dof
_map_onto_physical_1D = """\
// Evaluate basis functions for affine mapping
const double w0 = 1.0 - X_%(i)d[%(j)s][0];
const double w1 = X_%(i)d[%(j)s][0];

// Compute affine mapping y = F(X)
y[0] = w0*coordinate_dofs[0] + w1*coordinate_dofs[1];"""

_map_onto_physical_2D = """\
// Evaluate basis functions for affine mapping
const double w0 = 1.0 - X_%(i)d[%(j)s][0] - X_%(i)d[%(j)s][1];
const double w1 = X_%(i)d[%(j)s][0];
const double w2 = X_%(i)d[%(j)s][1];

// Compute affine mapping y = F(X)
y[0] = w0*coordinate_dofs[0] + w1*coordinate_dofs[2] + w2*coordinate_dofs[4];
y[1] = w0*coordinate_dofs[1] + w1*coordinate_dofs[3] + w2*coordinate_dofs[5];"""

_map_onto_physical_2D_1D = """\
// Evaluate basis functions for affine mapping
const double w0 = 1.0 - X_%(i)d[%(j)s][0];
const double w1 = X_%(i)d[%(j)s][0];

// Compute affine mapping y = F(X)
y[0] = w0*coordinate_dofs[0] + w1*coordinate_dofs[2];
y[1] = w0*coordinate_dofs[1] + w1*coordinate_dofs[3];"""

_map_onto_physical_3D = """\
// Evaluate basis functions for affine mapping
const double w0 = 1.0 - X_%(i)d[%(j)s][0] - X_%(i)d[%(j)s][1] - X_%(i)d[%(j)s][2];
const double w1 = X_%(i)d[%(j)s][0];
const double w2 = X_%(i)d[%(j)s][1];
const double w3 = X_%(i)d[%(j)s][2];

// Compute affine mapping y = F(X)
y[0] = w0*coordinate_dofs[0] + w1*coordinate_dofs[3] + w2*coordinate_dofs[6] + w3*coordinate_dofs[9];
y[1] = w0*coordinate_dofs[1] + w1*coordinate_dofs[4] + w2*coordinate_dofs[7] + w3*coordinate_dofs[10];
y[2] = w0*coordinate_dofs[2] + w1*coordinate_dofs[5] + w2*coordinate_dofs[8] + w3*coordinate_dofs[11];"""

_map_onto_physical_3D_1D = """\
// Evaluate basis functions for affine mapping
const double w0 = 1.0 - X_%(i)d[%(j)s][0];
const double w1 = X_%(i)d[%(j)s][0];

// Compute affine mapping y = F(X)
y[0] = w0*coordinate_dofs[0] + w1*coordinate_dofs[3];
y[1] = w0*coordinate_dofs[1] + w1*coordinate_dofs[4];
y[2] = w0*coordinate_dofs[2] + w1*coordinate_dofs[5];"""

_map_onto_physical_3D_2D = """\
// Evaluate basis functions for affine mapping
const double w0 = 1.0 - X_%(i)d[%(j)s][0] - X_%(i)d[%(j)s][1];
const double w1 = X_%(i)d[%(j)s][0];
const double w2 = X_%(i)d[%(j)s][1];

// Compute affine mapping y = F(X)
y[0] = w0*coordinate_dofs[0] + w1*coordinate_dofs[3] + w2*coordinate_dofs[6];
y[1] = w0*coordinate_dofs[1] + w1*coordinate_dofs[4] + w2*coordinate_dofs[7];
y[2] = w0*coordinate_dofs[2] + w1*coordinate_dofs[5] + w2*coordinate_dofs[8];
"""

_ip_coordinates_1D = """\
X%(num_ip)d[0] = %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[0] + \
                 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[1];"""

_ip_coordinates_2D_1D = """\
X%(num_ip)d[0] =\
 %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[0] +\
 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[2];
X%(num_ip)d[1] =\
 %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[1] +\
 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[3];"""

_ip_coordinates_3D_1D = """\
X%(num_ip)d[0] =\
 %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[0] +\
 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[3];
X%(num_ip)d[1] =\
 %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[1] +\
 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[4];
X%(num_ip)d[2] =\
 %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[2] +\
 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[5];"""

_ip_coordinates_2D = """\
X%(num_ip)d[0] = %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[0] + \
                 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[2] + %(name)s[%(ip)s][2]*coordinate_dofs%(restriction)s[4];
X%(num_ip)d[1] = %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[1] + \
                 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[3] + %(name)s[%(ip)s][2]*coordinate_dofs%(restriction)s[5];"""

_ip_coordinates_3D_2D = """\
X%(num_ip)d[0] =\
 %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[0] +\
 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[3] +\
 %(name)s[%(ip)s][2]*coordinate_dofs%(restriction)s[6];
X%(num_ip)d[1] =\
 %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[1] +\
 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[4] +\
 %(name)s[%(ip)s][2]*coordinate_dofs%(restriction)s[7];
X%(num_ip)d[2] =\
 %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[2] +\
 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[5] +\
 %(name)s[%(ip)s][2]*coordinate_dofs%(restriction)s[8];"""

_ip_coordinates_3D = """\
X%(num_ip)d[0] = %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[0] + \
                 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[3] + \
                 %(name)s[%(ip)s][2]*coordinate_dofs%(restriction)s[6] + \
                 %(name)s[%(ip)s][3]*coordinate_dofs%(restriction)s[9];
X%(num_ip)d[1] = %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[1] + \
                 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[4] + \
                 %(name)s[%(ip)s][2]*coordinate_dofs%(restriction)s[7] + \
                 %(name)s[%(ip)s][3]*coordinate_dofs%(restriction)s[10];
X%(num_ip)d[2] = %(name)s[%(ip)s][0]*coordinate_dofs%(restriction)s[2] + \
                 %(name)s[%(ip)s][1]*coordinate_dofs%(restriction)s[5] + \
                 %(name)s[%(ip)s][2]*coordinate_dofs%(restriction)s[8] + \
                 %(name)s[%(ip)s][3]*coordinate_dofs%(restriction)s[11];"""

# Codesnippets used in evaluatebasis[|derivatives]
_map_coordinates_FIAT_interval = """\
// Get coordinates and map to the reference (FIAT) element
double X = (2.0*x[0] - coordinate_dofs[0] - coordinate_dofs[1]) / J[0];"""

_map_coordinates_FIAT_interval_in_2D = """\
// Get coordinates and map to the reference (FIAT) element
double X = 2*(std::sqrt(std::pow(x[0] - coordinate_dofs[0], 2) + std::pow(x[1] - coordinate_dofs[1], 2)) / detJ) - 1.0;"""

_map_coordinates_FIAT_interval_in_3D = """\
// Get coordinates and map to the reference (FIAT) element
double X = 2*(std::sqrt(std::pow(x[0] - coordinate_dofs[0], 2) + std::pow(x[1] - coordinate_dofs[1], 2) + std::pow(x[2] - coordinate_dofs[2], 2))/ detJ) - 1.0;"""

_map_coordinates_FIAT_triangle = """\
// Compute constants
const double C0 = coordinate_dofs[2] + coordinate_dofs[4];
const double C1 = coordinate_dofs[3] + coordinate_dofs[5];

// Get coordinates and map to the reference (FIAT) element
double X = (J[1]*(C1 - 2.0*x[1]) + J[3]*(2.0*x[0] - C0)) / detJ;
double Y = (J[0]*(2.0*x[1] - C1) + J[2]*(C0 - 2.0*x[0])) / detJ;"""

_map_coordinates_FIAT_triangle_in_3D = """\
const double b0 = coordinate_dofs[0];
const double b1 = coordinate_dofs[1];
const double b2 = coordinate_dofs[2];

// P_FFC = J^dag (p - b), P_FIAT = 2*P_FFC - (1, 1)
double X = 2*(K[0]*(x[0] - b0) + K[1]*(x[1] - b1) + K[2]*(x[2] - b2)) - 1.0;
double Y = 2*(K[3]*(x[0] - b0) + K[4]*(x[1] - b1) + K[5]*(x[2] - b2)) - 1.0;
"""

_map_coordinates_FIAT_tetrahedron = """\
// Compute constants
const double C0 = coordinate_dofs[9]  + coordinate_dofs[6] + coordinate_dofs[3]  - coordinate_dofs[0];
const double C1 = coordinate_dofs[10] + coordinate_dofs[7] + coordinate_dofs[4]  - coordinate_dofs[1];
const double C2 = coordinate_dofs[11] + coordinate_dofs[8] + coordinate_dofs[5]  - coordinate_dofs[2];

// Compute subdeterminants
const double d_00 = J[4]*J[8] - J[5]*J[7];
const double d_01 = J[5]*J[6] - J[3]*J[8];
const double d_02 = J[3]*J[7] - J[4]*J[6];
const double d_10 = J[2]*J[7] - J[1]*J[8];
const double d_11 = J[0]*J[8] - J[2]*J[6];
const double d_12 = J[1]*J[6] - J[0]*J[7];
const double d_20 = J[1]*J[5] - J[2]*J[4];
const double d_21 = J[2]*J[3] - J[0]*J[5];
const double d_22 = J[0]*J[4] - J[1]*J[3];

// Get coordinates and map to the reference (FIAT) element
double X = (d_00*(2.0*x[0] - C0) + d_10*(2.0*x[1] - C1) + d_20*(2.0*x[2] - C2)) / detJ;
double Y = (d_01*(2.0*x[0] - C0) + d_11*(2.0*x[1] - C1) + d_21*(2.0*x[2] - C2)) / detJ;
double Z = (d_02*(2.0*x[0] - C0) + d_12*(2.0*x[1] - C1) + d_22*(2.0*x[2] - C2)) / detJ;
"""

# Mappings to code snippets used by format These dictionaries accept
# as keys: first the topological dimension, and second the geometric
# dimension

facet_determinant = {1: {1: _facet_determinant_1D,
                         2: _facet_determinant_2D_1D,
                         3: _facet_determinant_3D_1D},
                     2: {2: _facet_determinant_2D,
                         3: _facet_determinant_3D_2D},
                     3: {3: _facet_determinant_3D}}

# Geometry related snippets
map_onto_physical = {1: {1: _map_onto_physical_1D,
                         2: _map_onto_physical_2D_1D,
                         3: _map_onto_physical_3D_1D},
                     2: {2: _map_onto_physical_2D,
                         3: _map_onto_physical_3D_2D},
                     3: {3: _map_onto_physical_3D}}

fiat_coordinate_map = {"interval": {1: _map_coordinates_FIAT_interval,
                                    2: _map_coordinates_FIAT_interval_in_2D,
                                    3: _map_coordinates_FIAT_interval_in_3D},
                       "triangle": {2: _map_coordinates_FIAT_triangle,
                                    3: _map_coordinates_FIAT_triangle_in_3D},
                       "tetrahedron": {3: _map_coordinates_FIAT_tetrahedron}}

transform_snippet = {"interval": {1: _transform_snippet(1, 1),
                                  2: _transform_snippet(1, 2),
                                  3: _transform_snippet(1, 3)},
                     "triangle": {2: _transform_snippet(2, 2),
                                  3: _transform_snippet(2, 3)},
                     "tetrahedron": {3: _transform_snippet(3, 3)}}

ip_coordinates = {1: {1: (3, _ip_coordinates_1D),
                      2: (6, _ip_coordinates_2D_1D),
                      3: (9, _ip_coordinates_3D_1D)},
                  2: {2: (10, _ip_coordinates_2D),
                      3: (15, _ip_coordinates_3D_2D)},
                  3: {3: (21, _ip_coordinates_3D)}}

# FIXME: Rename as in compute_jacobian _compute_foo_<shape>_<n>d

normal_direction = {1: {1: _normal_direction_1D,
                        2: _normal_direction_2D_1D,
                        3: _normal_direction_3D_1D},
                    2: {2: _normal_direction_2D,
                        3: _normal_direction_3D_2D},
                    3: {3: _normal_direction_3D}}

facet_normal = {1: {1: _facet_normal_1D,
                    2: _facet_normal_2D_1D,
                    3: _facet_normal_3D_1D},
                2: {2: _facet_normal_2D,
                    3: _facet_normal_3D_2D},
                3: {3: _facet_normal_3D}}

cell_volume = {1: {1: _cell_volume_1D,
                   2: _cell_volume_2D_1D,
                   3: _cell_volume_3D_1D},
               2: {2: _cell_volume_2D,
                   3: _cell_volume_3D_2D},
               3: {3: _cell_volume_3D}}

circumradius = {1: {1: _circumradius_1D,
                    2: _circumradius_2D_1D,
                    3: _circumradius_3D_1D},
                2: {2: _circumradius_2D,
                    3: _circumradius_3D_2D},
                3: {3: _circumradius_3D}}

facet_area = {1: {1: _facet_area_1D,
                  2: _facet_area_2D_1D,
                  3: _facet_area_3D_1D},
              2: {2: _facet_area_2D,
                  3: _facet_area_3D_2D},
              3: {3: _facet_area_3D}}

min_facet_edge_length = {3: {3: _min_facet_edge_length_3D}}

max_facet_edge_length = {3: {3: _max_facet_edge_length_3D}}

# Code snippets for runtime quadrature (calling evaluate_basis)

eval_basis_decl = """\
std::vector<std::vector<double> > %(table_name)s(num_quadrature_points);"""

eval_basis_init = """\
for (std::size_t ip = 0; ip < num_quadrature_points; ip++)
  %(table_name)s[ip].resize(%(table_size)s);"""

eval_basis = """\
// Get current quadrature point and compute values of basis functions
const double* x = quadrature_points + ip*%(gdim)s;
const double* v = coordinate_dofs + %(vertex_offset)s;
%(classname)s::_evaluate_basis_all(%(eval_name)s, x, v, cell_orientation);"""

eval_basis_copy = """\
// Copy values to table %(table_name)s
for (std::size_t i = 0; i < %(space_dim)s; i++)
  %(table_name)s[ip][%(table_offset)s + i] = %(eval_name)s[%(eval_stride)s*i + %(eval_offset)s];"""

eval_derivs_decl = """\
std::vector<std::vector<double> > %(table_name)s(num_quadrature_points);"""

eval_derivs_init = """\
for (std::size_t ip = 0; ip < num_quadrature_points; ip++)
  %(table_name)s[ip].resize(%(table_size)s);"""

eval_derivs = """\
// Get current quadrature point and compute values of basis function derivatives
const double* x = quadrature_points + ip*%(gdim)s;
const double* v = coordinate_dofs + %(vertex_offset)s;
%(classname)s::_evaluate_basis_derivatives_all(%(n)s, %(eval_name)s, x, v, cell_orientation);"""

eval_derivs_copy = """\
// Copy values to table %(table_name)s
for (std::size_t i = 0; i < %(space_dim)s; i++)
  %(table_name)s[ip][%(table_offset)s + i] = %(eval_name)s[%(eval_stride)s*i + %(eval_offset)s];"""



# -*- coding: utf-8 -*-

# Copyright (C) 2007-2016 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Anders Logg 2013
# Modified by Lizao Li 2015, 2016

"""Code generation for evaluation of derivatives of finite element
basis values.  This module generates code which is more or less a C++
representation of the code found in FIAT_NEW."""

# Python modules
import math
import numpy
from six import string_types

# FFC modules
from ffc.log import error
from ffc.evaluatebasis import _compute_basisvalues, _tabulate_coefficients
from ffc.cpp import remove_unused, indent, format


def _evaluate_basis_derivatives_all(data):
    """Like evaluate_basis, but return the values of all basis
    functions (dofs)."""

    if isinstance(data, string_types):
        return format["exception"]("evaluate_basis_derivatives_all: %s" % data)

    # Prefetch formats.
    f_r, f_s = format["free indices"][:2]
    f_assign = format["assign"]
    f_loop = format["generate loop"]
    f_array = format["dynamic array"]
    f_dof_vals = format["dof values"]
    f_comment = format["comment"]
    f_derivs = format["call basis_derivatives"]
    f_values = format["argument values"]
    f_int = format["int"]
    f_num_derivs = format["num derivatives"]
    f_double = format["float declaration"]
    f_component = format["component"]
    f_mul = format["mul"]
    f_float = format["floating point"]
    f_index = format["matrix index"]
    f_del_array = format["delete dynamic array"]

    # Initialise return code
    code = []

    # FIXME: KBO: Figure out which return format to use, either:
    # [dN0[0]/dx, dN0[0]/dy, dN0[1]/dx, dN0[1]/dy, dN1[0]/dx,
    # dN1[0]/dy, dN1[1]/dx, dN1[1]/dy, ...]
    # or
    # [dN0[0]/dx, dN1[0]/dx, ..., dN0[1]/dx, dN1[1]/dx, ...,
    # dN0[0]/dy, dN1[0]/dy, ..., dN0[1]/dy, dN1[1]/dy, ...]
    # or
    # [dN0[0]/dx, dN0[1]/dx, ..., dN1[0]/dx, dN1[1]/dx, ...,
    # dN0[0]/dy, dN0[1]/dy, ..., dN1[0]/dy, dN1[1]/dy, ...]
    # for vector (tensor elements), currently returning option 1.

    # FIXME: KBO: For now, just call evaluate_basis_derivatives and
    # map values accordingly, this will keep the amount of code at a
    # minimum. If it turns out that speed is an issue (overhead from
    # calling evaluate_basis), we can easily generate all the code.

    # Get total value shape and space dimension for entire element
    # (possibly mixed).
    physical_value_size = data["physical_value_size"]
    space_dimension = data["space_dimension"]
    max_degree = data["max_degree"]

    # Special case where space dimension is one (constant elements).
    if space_dimension == 1:
        code += [f_comment("Element is constant, calling evaluate_basis_derivatives.")]
        code += [f_derivs(f_int(0), f_values)]
        return "\n".join(code)

    # Compute number of derivatives.
    if data["topological_dimension"] == data["geometric_dimension"]:
        _g = ""
    else:
        _g = "_g"

    # If n == 0, call evaluate_basis.
    code += [format["comment"]("Call evaluate_basis_all if order of derivatives is equal to zero.")]
    cond = format["argument derivative order"] + format["is equal"] + format["int"](0)
    val = [format["call basis_all"]]
    val += [format["return"]("")]
    code += [format["if"](cond, indent("\n".join(val), 2))]

    code += _compute_num_derivatives(data["geometric_dimension"], _g)

    if (physical_value_size == 1):
        num_vals = f_num_derivs(_g)
    else:
        num_vals = f_mul([f_int(physical_value_size), f_num_derivs(_g)])

    # Reset values.
    code += ["", f_comment("Set values equal to zero.")]
    name = f_component(f_values, f_index(f_r, f_s, num_vals))
    lines_s = [f_assign(name, f_float(0.0))]
    loop_s = [(f_s, 0, num_vals)]
    lines_r = f_loop(lines_s, loop_s)
    loop_r = [(f_r, 0, space_dimension)]
    code += f_loop(lines_r, loop_r)

    # If n > max_degree, return zeros.
    code += ["", format["comment"]("If order of derivatives is greater than the maximum polynomial degree, return zeros.")]
    cond = format["argument derivative order"] + format["greater than"] + f_int(max_degree)
    val = format["return"]("")
    code += [format["if"](cond, indent(val, 2))]

    # Declare helper value to hold single dof values and reset.
    code += [f_comment("Helper variable to hold values of a single dof.")]
    nds = data["geometric_dimension"]**max_degree * physical_value_size
    code += [format["declaration"](f_double, f_component(f_dof_vals, f_int(nds)))]
    line = [f_assign(f_component(f_dof_vals, f_r), f_float(0.0))]
    code += f_loop(line, [(f_r, 0, nds)])

    # Create loop over dofs that calls evaluate_basis_derivatives for a single dof and
    # inserts the values into the global array.
    code += ["", f_comment("Loop dofs and call evaluate_basis_derivatives.")]
    name = f_component(f_values, f_index(f_r, f_s, num_vals))
    value = f_component(f_dof_vals, f_s)
    lines_s = [f_assign(name, value)]
    loop_s = [(f_s, 0, num_vals)]

    lines_r = [f_derivs(f_r, f_dof_vals)]
    lines_r += f_loop(lines_s, loop_s)
    loop_r = [(f_r, 0, space_dimension)]
    code += f_loop(lines_r, loop_r)

    # Generate bode (no need to remove unused).
    return "\n".join(code)


def _evaluate_basis_derivatives(data):
    """Evaluate the derivatives of an element basisfunction at a point. The values are
    computed as in FIAT as the matrix product of the coefficients (computed at compile time),
    basisvalues which are dependent on the coordinate and thus have to be computed at
    run time and combinations (depending on the order of derivative) of dmats
    tables which hold the derivatives of the expansion coefficients."""

    if isinstance(data, string_types):
        return format["exception"]("evaluate_basis_derivatives: %s" % data)

    # Initialise return code.
    code = []

    # Get the element cell domain, geometric and topological dimension.
    element_cellname = data["cellname"]
    gdim = data["geometric_dimension"]
    tdim = data["topological_dimension"]
    max_degree = data["max_degree"]

    # Compute number of derivatives that has to be computed, and
    # declare an array to hold the values of the derivatives on the
    # reference element.
    code += [""]
    if tdim == gdim:
        _t = ""
        _g = ""
        code += _compute_num_derivatives(tdim, "")

        # Reset all values.
        code += _reset_values(data, _g)

        # Handle values of argument 'n'.
        code += _handle_degree(max_degree)

        # If max_degree is zero, return code (to avoid declarations such as
        # combinations[1][0]) and because there's nothing to compute.)
        if max_degree == 0:
            return remove_unused("\n".join(code))

        # Generate geo code.
        code += _geometry_related_code(data, tdim, gdim, element_cellname)

        # Generate all possible combinations of derivatives.
        code += _generate_combinations(tdim, "", max_degree)
    else:
        _t = "_t"
        _g = "_g"
        code += _compute_num_derivatives(tdim, _t)
        code += [""]
        code += _compute_num_derivatives(gdim, _g)

        # Reset all values.
        code += _reset_values(data, _g)

        # Handle values of argument 'n'.
        code += _handle_degree(max_degree)

        # If max_degree is zero, return code (to avoid declarations such as
        # combinations[1][0]) and because there's nothing to compute.)
        if max_degree == 0:
            return remove_unused("\n".join(code))

        # Generate geo code.
        code += _geometry_related_code(data, tdim, gdim, element_cellname)

        # Generate all possible combinations of derivatives.
        code += _generate_combinations(tdim, _t, max_degree)
        code += _generate_combinations(gdim, _g, max_degree)

    # Generate the transformation matrix.
    code += _generate_transform(element_cellname, gdim, tdim, max_degree)

    # Create code for all basis values (dofs).
    dof_cases = []
    for dof_data in data["dofs_data"]:
        dof_cases.append(_generate_dof_code(data, dof_data))
    code += [format["switch"](format["argument basis num"], dof_cases)]
    code = remove_unused("\n".join(code))
    # code = "\n".join(code)
    return code


def _handle_degree(max_degree):
    """Check value of argument 'n' against the maximum polynomial degree of the
    finite element. If user ask for n>max_degree return an appropriate number
    of zeros in the 'values' array. If n==0, simply direct call to
    evaluate_basis."""

    code = []

    # If n == 0, call evaluate_basis.
    code += [format["comment"]("Call evaluate_basis if order of derivatives is equal to zero.")]
    cond = format["argument derivative order"] + format["is equal"] + format["int"](0)
    val = [format["call basis"](format["argument dof num"], format["argument values"])]
    val += [format["return"]("")]
    code += [format["if"](cond, indent("\n".join(val), 2))]

    # If n > max_degree, derivatives are always zero. Since the appropriate number of
    # zeros have already been inserted into the 'values' array simply return.
    code += [format["comment"]("If order of derivatives is greater than the maximum polynomial degree, return zeros.")]
    cond = format["argument derivative order"] + format["greater than"] + format["int"](max_degree)
    val = format["return"]("")
    code += [format["if"](cond, val)]

    return code


def _geometry_related_code(data, tdim, gdim, element_cellname):
    code = []
    # Get code snippets for Jacobian, inverse of Jacobian and mapping of
    # coordinates from physical element to the FIAT reference element.
    code += [format["compute_jacobian"](tdim, gdim)]
    code += [format["compute_jacobian_inverse"](tdim, gdim)]
    if data["needs_oriented"]:
        code += [format["orientation"](tdim, gdim)]
    code += ["", format["fiat coordinate map"](element_cellname, gdim)]
    return code


def _compute_num_derivatives(dimension, suffix=""):
    """Computes the number of derivatives of order 'n' as dimension()^n.

    Dimension will be the element topological dimension for the number
    of derivatives in local coordinates, and the geometric dimension
    for the number of derivatives in phyisical coordinates.
    """
    # Prefetch formats.
    f_int = format["int"]
    f_num_derivs = format["num derivatives"](suffix)

    # Use loop to compute power since using std::pow() result in an
    # ambiguous call.
    code = [format["comment"]("Compute number of derivatives.")]
    code.append(format["declaration"](format["uint declaration"],
                                      f_num_derivs, f_int(1)))
    loop_vars = [(format["free indices"][0], 0,
                  format["argument derivative order"])]
    lines = [format["imul"](f_num_derivs, f_int(dimension))]
    code += format["generate loop"](lines, loop_vars)

    return code


def _generate_combinations(dimension, suffix, max_degree):
    "Generate all possible combinations of derivatives of order 'n'."

    nds = dimension**max_degree

    # Use code from format.
    code = ["", format["combinations"]
            % {"combinations": format["derivative combinations"](suffix),
               "dimension-1": dimension - 1,
               "num_derivatives": format["num derivatives"](suffix),
               "n": format["argument derivative order"],
               "max_num_derivatives":format["int"](nds),
               "max_degree":format["int"](max_degree)}]
    return code


def _generate_transform(element_cellname, gdim, tdim, max_degree):
    """Generate the transformation matrix, which is used to transform
    derivatives from reference element back to the physical element."""

    max_g_d = gdim**max_degree
    max_t_d = tdim**max_degree
    # Generate code to construct the inverse of the Jacobian
    if (element_cellname in ["interval", "triangle", "tetrahedron"]):
        code = ["", format["transform snippet"][element_cellname][gdim]
                % {"transform": format["transform matrix"],
                   "num_derivatives": format["num derivatives"](""),
                   "n": format["argument derivative order"],
                   "combinations": format["derivative combinations"](""),
                   "K":format["transform Jinv"],
                   "max_g_deriv":max_g_d, "max_t_deriv":max_t_d}]
    else:
        error("Cannot generate transform for shape: %s" % element_cellname)

    return code


def _reset_values(data, suffix):
    "Reset all components of the 'values' array as it is a pointer to an array."

    # Prefetch formats.
    f_assign = format["assign"]
    f_r = format["free indices"][0]

    code = ["", format["comment"]("Reset values. Assuming that values is always an array.")]

    # Get value shape and reset values. This should also work for TensorElement,
    # scalar are empty tuples, therefore (1,) in which case value_shape = 1.
    physical_value_size = data["physical_value_size"]

    # Only multiply by value shape if different from 1.
    if physical_value_size == 1:
        num_vals = format["num derivatives"](suffix)
    else:
        num_vals = format["mul"]([format["int"](physical_value_size), format["num derivatives"](suffix)])
    name = format["component"](format["argument values"], f_r)
    loop_vars = [(f_r, 0, num_vals)]
    lines = [f_assign(name, format["floating point"](0))]
    code += format["generate loop"](lines, loop_vars)

    return code + [""]


def _generate_dof_code(data, dof_data):
    "Generate code for a basis."

    code = []

    # Compute basisvalues, from evaluatebasis.py.
    code += _compute_basisvalues(data, dof_data)

    # Tabulate coefficients.
    code += _tabulate_coefficients(dof_data)

    # Tabulate coefficients for derivatives.
    code += _tabulate_dmats(dof_data)

    # Compute the derivatives of the basisfunctions on the reference (FIAT) element,
    # as the dot product of the new coefficients and basisvalues.
    code += _compute_reference_derivatives(data, dof_data)

    # Transform derivatives to physical element by multiplication with the transformation matrix.
    code += _transform_derivatives(data, dof_data)

    code = remove_unused("\n".join(code))

    return code


def _tabulate_dmats(dof_data):
    "Tabulate the derivatives of the polynomial base"

    code = []

    # Prefetch formats to speed up code generation.
    f_table = format["static const float declaration"]
    f_dmats = format["dmats"]
    f_component = format["component"]
    f_decl = format["declaration"]
    f_tensor = format["tabulate tensor"]
    f_new_line = format["new line"]

    # Get derivative matrices (coefficients) of basis functions, computed by FIAT at compile time.
    derivative_matrices = dof_data["dmats"]

    code += [format["comment"]("Tables of derivatives of the polynomial base (transpose).")]

    # Generate tables for each spatial direction.
    for i, dmat in enumerate(derivative_matrices):

        # Extract derivatives for current direction (take transpose, FIAT_NEW PolynomialSet.tabulate()).
        matrix = numpy.transpose(dmat)

        # Get shape and check dimension (This is probably not needed).
        shape = numpy.shape(matrix)
        if not (shape[0] == shape[1] == dof_data["num_expansion_members"]):
            error("Something is wrong with the shape of dmats.")

        # Declare varable name for coefficients.
        name = f_component(f_dmats(i), [shape[0], shape[1]])
        code += [f_decl(f_table, name, f_new_line + f_tensor(matrix)), ""]

    return code


def _reset_dmats(shape_dmats, indices):
    "Set values in dmats equal to the identity matrix."
    f_assign = format["assign"]
    f_float = format["floating point"]
    i, j = indices

    code = [format["comment"]("Resetting dmats values to compute next derivative.")]
    dmats_old = format["component"](format["dmats"](""), [i, j])
    lines = [f_assign(dmats_old, f_float(0.0))]
    lines += [format["if"](i + format["is equal"] + j,
              f_assign(dmats_old, f_float(1.0)))]
    loop_vars = [(i, 0, shape_dmats[0]), (j, 0, shape_dmats[1])]
    code += format["generate loop"](lines, loop_vars)
    return code


def _update_dmats(shape_dmats, indices):
    "Update values in dmats_old with values in dmats and set values in dmats to zero."
    f_assign = format["assign"]
    f_component = format["component"]
    i, j = indices

    code = [format["comment"]("Updating dmats_old with new values and resetting dmats.")]
    dmats = f_component(format["dmats"](""), [i, j])
    dmats_old = f_component(format["dmats old"], [i, j])
    lines = [f_assign(dmats_old, dmats), f_assign(dmats, format["floating point"](0.0))]
    loop_vars = [(i, 0, shape_dmats[0]), (j, 0, shape_dmats[1])]
    code += format["generate loop"](lines, loop_vars)
    return code


def _compute_dmats(num_dmats, shape_dmats, available_indices, deriv_index, _t):
    "Compute values of dmats as a matrix product."
    f_comment = format["comment"]
    s, t, u = available_indices

    # Reset dmats_old
    code = _reset_dmats(shape_dmats, [t, u])
    code += ["", f_comment("Looping derivative order to generate dmats.")]

    # Set dmats matrix equal to dmats_old
    lines = _update_dmats(shape_dmats, [t, u])

    lines += ["", f_comment("Update dmats using an inner product.")]

    # Create dmats matrix by multiplication
    comb = format["component"](format["derivative combinations"](_t), [deriv_index, s])
    for i in range(num_dmats):
        lines += _dmats_product(shape_dmats, comb, i, [t, u])

    loop_vars = [(s, 0, format["argument derivative order"])]
    code += format["generate loop"](lines, loop_vars)

    return code


def _dmats_product(shape_dmats, index, i, indices):
    "Create product to update dmats."
    f_loop = format["generate loop"]
    f_component = format["component"]
    t, u = indices
    tu = t + u

    dmats = f_component(format["dmats"](""), [t, u])
    dmats_old = f_component(format["dmats old"], [tu, u])
    value = format["multiply"]([f_component(format["dmats"](i), [t, tu]), dmats_old])
    name = format["iadd"](dmats, value)
    lines = f_loop([name], [(tu, 0, shape_dmats[0])])
    loop_vars = [(t, 0, shape_dmats[0]), (u, 0, shape_dmats[1])]
    code = [format["if"](index + format["is equal"] + str(i),
            "\n".join(f_loop(lines, loop_vars)))]
    return code


def _compute_reference_derivatives(data, dof_data):
    """Compute derivatives on the reference element by recursively multiply coefficients with
    the relevant derivatives of the polynomial base until the requested order of derivatives
    has been reached. After this take the dot product with the basisvalues."""

    # Prefetch formats to speed up code generation
    f_comment = format["comment"]
    f_num_derivs = format["num derivatives"]
    f_mul = format["mul"]
    f_int = format["int"]
    f_matrix_index = format["matrix index"]
    f_coefficients = format["coefficients"]
#    f_dof           = format["local dof"]
    f_basisvalues = format["basisvalues"]
    f_const_double = format["const float declaration"]
    f_group = format["grouping"]
    f_transform = format["transform"]
    f_double = format["float declaration"]
    f_component = format["component"]
    f_tmp = format["tmp ref value"]
    f_dmats = format["dmats"]
    f_dmats_old = format["dmats old"]
    f_assign = format["assign"]
    f_decl = format["declaration"]
    f_iadd = format["iadd"]
    f_add = format["add"]
    f_tensor = format["tabulate tensor"]
    f_new_line = format["new line"]
    f_loop = format["generate loop"]
    f_derivatives = format["reference derivatives"]
    f_array = format["dynamic array"]
    f_float = format["floating point"]
    f_inv = format["inverse"]
    f_detJ = format["det(J)"]
    f_inner = format["inner product"]

    f_r, f_s, f_t, f_u = format["free indices"]

    tdim = data["topological_dimension"]
    gdim = data["geometric_dimension"]
    max_degree = data["max_degree"]

    if tdim == gdim:
        _t = ""
        _g = ""
    else:
        _t = "_t"
        _g = "_g"

    # Get number of components.
    num_components = dof_data["num_components"]

    # Get shape of derivative matrix (they should all have the same shape) and
    # verify that it is a square matrix.
    shape_dmats = numpy.shape(dof_data["dmats"][0])
    if shape_dmats[0] != shape_dmats[1]:
        error("Something is wrong with the dmats:\n%s" % str(dof_data["dmats"]))

    code = [f_comment("Compute reference derivatives.")]

    # Declare pointer to array that holds derivatives on the FIAT element
    code += [f_comment("Declare array of derivatives on FIAT element.")]
    # The size of the array of reference derivatives is equal to the number of derivatives
    # times the number of components of the basis element
    if (num_components == 1):
        num_vals = f_num_derivs(_t)
    else:
        num_vals = f_mul([f_int(num_components), f_num_derivs(_t)])

    nds = tdim**max_degree * num_components
    code += [format["declaration"](f_double, f_component(f_derivatives, f_int(nds)))]
    line = [f_assign(f_component(f_derivatives, f_r), f_float(0.0))]
    code += f_loop(line, [(f_r, 0, nds)])
    code += [""]

    mapping = dof_data["mapping"]
    if "piola" in mapping:
        # In either of the Piola cases, the value space of the derivatives is the geometric dimension rather than the topological dimension.
        code += [f_comment("Declare array of reference derivatives on physical element.")]

        _p = "_p"
        num_components_p = gdim

        nds = tdim**max_degree * gdim
        code += [format["declaration"](f_double, f_component(f_derivatives + _p, f_int(nds)))]
        line = [f_assign(f_component(f_derivatives + _p, f_r), f_float(0.0))]
        code += f_loop(line, [(f_r, 0, nds)])
        code += [""]
    else:
        _p = ""
        num_components_p = num_components

    # Declare matrix of dmats (which will hold the matrix product of all combinations)
    # and dmats_old which is needed in order to perform the matrix product.
    value = f_tensor(numpy.eye(shape_dmats[0]))
    code += [f_comment("Declare derivative matrix (of polynomial basis).")]
    name = f_component(f_dmats(""), [shape_dmats[0], shape_dmats[1]])
    code += [f_decl(f_double, name, f_new_line + value), ""]
    code += [f_comment("Declare (auxiliary) derivative matrix (of polynomial basis).")]
    name = f_component(f_dmats_old, [shape_dmats[0], shape_dmats[1]])
    code += [f_decl(f_double, name, f_new_line + value), ""]

    # Compute dmats as a recursive matrix product
    lines = _compute_dmats(len(dof_data["dmats"]), shape_dmats, [f_s, f_t, f_u], f_r, _t)

    # Compute derivatives for all components
    lines_c = []
    for i in range(num_components):
        name = f_component(f_derivatives, f_matrix_index(i, f_r, f_num_derivs(_t)))
        coeffs = f_component(f_coefficients(i), f_s)
        dmats = f_component(f_dmats(""), [f_s, f_t])
        basis = f_component(f_basisvalues, f_t)
        lines_c.append(f_iadd(name, f_mul([coeffs, dmats, basis])))
    loop_vars_c = [(f_s, 0, shape_dmats[0]), (f_t, 0, shape_dmats[1])]
    lines += f_loop(lines_c, loop_vars_c)

    # Apply transformation if applicable.
    if mapping == "affine":
        pass
    elif mapping == "contravariant piola":
        lines += ["", f_comment
                  ("Using contravariant Piola transform to map values back to the physical element.")]
        # Get temporary values before mapping.
        lines += [f_const_double(f_tmp(i),
                  f_component(f_derivatives, f_matrix_index(i, f_r, f_num_derivs(_t)))) for i in range(num_components)]

        # Create names for inner product.
        basis_col = [f_tmp(j) for j in range(tdim)]
        for i in range(num_components_p):
            # Create Jacobian.
            jacobian_row = [f_transform("J", i, j, gdim, tdim, None) for j in range(tdim)]

            # Create inner product and multiply by inverse of Jacobian.
            inner = [f_mul([jacobian_row[j], basis_col[j]]) for j in range(tdim)]
            sum_ = f_group(f_add(inner))
            value = f_mul([f_inv(f_detJ(None)), sum_])
            name = f_component(f_derivatives + _p, f_matrix_index(i, f_r, f_num_derivs(_t)))
            lines += [f_assign(name, value)]
    elif mapping == "covariant piola":
        lines += ["", f_comment
                  ("Using covariant Piola transform to map values back to the physical element")]
        # Get temporary values before mapping.
        lines += [f_const_double(f_tmp(i),
                  f_component(f_derivatives, f_matrix_index(i, f_r, f_num_derivs(_t)))) for i in range(num_components)]
        # Create names for inner product.
        basis_col = [f_tmp(j) for j in range(tdim)]
        for i in range(num_components_p):
            # Create inverse of Jacobian.
            inv_jacobian_column = [f_transform("JINV", j, i, tdim, gdim, None) for j in range(tdim)]

            # Create inner product of basis and inverse of Jacobian.
            inner = [f_mul([inv_jacobian_column[j], basis_col[j]]) for j in range(tdim)]
            value = f_group(f_add(inner))
            name = f_component(f_derivatives + _p, f_matrix_index(i, f_r, f_num_derivs(_t)))
            lines += [f_assign(name, value)]
    elif mapping == "double covariant piola":
        code += ["", f_comment("Using double covariant Piola transform to map values back to the physical element")]
        lines += [f_const_double(f_tmp(i),
                                 f_component(f_derivatives,
                                             f_matrix_index(i, f_r, f_num_derivs(_t))))
                  for i in range(num_components)]
        basis_col = [f_tmp(j) for j in range(num_components)]
        for p in range(num_components):
            # unflatten the indices
            i = p // tdim
            l = p % tdim
            # g_il = K_ji G_jk K_kl
            value = f_group(f_inner(
                [f_inner([f_transform("JINV", j, i, tdim, gdim, None)
                          for j in range(tdim)],
                         [basis_col[j * tdim + k] for j in range(tdim)])
                 for k in range(tdim)],
                [f_transform("JINV", k, l, tdim, gdim, None)
                 for k in range(tdim)]))
            name = f_component(f_derivatives + _p, f_matrix_index(p, f_r, f_num_derivs(_t)))
            lines += [f_assign(name, value)]
    elif mapping == "double contravariant piola":
        code += ["", f_comment("Using double contravariant Piola transform to map values back to the physical element.")]
        lines += [f_const_double(
            f_tmp(i),
            f_component(f_derivatives,
                        f_matrix_index(i, f_r, f_num_derivs(_t))))
                  for i in range(num_components)]
        basis_col = [f_tmp(j) for j in range(num_components)]
        for p in range(num_components):
            # unflatten the indices
            i = p // tdim
            l = p % tdim
            # g_il = (det J)^(-2) Jij G_jk Jlk
            value = f_group(f_inner(
                [f_inner([f_transform("J", i, j, tdim, gdim, None)
                          for j in range(tdim)],
                         [basis_col[j * tdim + k] for j in range(tdim)])
                 for k in range(tdim)],
                [f_transform("J", l, k, tdim, gdim, None)
                 for k in range(tdim)]))
            value = f_mul([f_inv(f_detJ(None)), f_inv(f_detJ(None)), value])
            name = f_component(f_derivatives+_p,
                               f_matrix_index(p, f_r, f_num_derivs(_t)))
            lines += [f_assign(name, value)]
    else:
        error("Unknown mapping: %s" % mapping)

    # Generate loop over number of derivatives.
    # Loop all derivatives and compute value of the derivative as:
    # deriv_on_ref[r] = coeff[dof][s]*dmat[s][t]*basis[t]
    code += [f_comment("Loop possible derivatives.")]
    loop_vars = [(f_r, 0, f_num_derivs(_t))]
    code += f_loop(lines, loop_vars)

    return code + [""]


def _transform_derivatives(data, dof_data):
    """Transform derivatives back to the physical element by applying the
    transformation matrix."""

    # Prefetch formats to speed up code generation.
    f_loop = format["generate loop"]
    f_num_derivs = format["num derivatives"]
    f_derivatives = format["reference derivatives"]
    f_values = format["argument values"]
    f_mul = format["mul"]
    f_iadd = format["iadd"]
    f_component = format["component"]
    f_transform = format["transform matrix"]
    f_r, f_s = format["free indices"][:2]
    f_index = format["matrix index"]

    if data["topological_dimension"] == data["geometric_dimension"]:
        _t = ""
        _g = ""
    else:
        _t = "_t"
        _g = "_g"

    # Get number of components and offset.
    num_components = dof_data["num_components"]
    reference_offset = dof_data["reference_offset"]
    physical_offset = dof_data["physical_offset"]
    offset = reference_offset  # physical_offset # FIXME: Should be physical offset but that breaks tests

    mapping = dof_data["mapping"]
    if "piola" in mapping:
        # In either of the Piola cases, the value space of the derivatives is the geometric dimension rather than the topological dimension.
        _p = "_p"
        num_components_p = data["geometric_dimension"]
    else:
        _p = ""
        num_components_p = num_components

    code = [format["comment"]("Transform derivatives back to physical element")]

    lines = []
    for i in range(num_components_p):
        access_name = f_index(offset + i, f_r, f_num_derivs(_g))
        name = f_component(f_values, access_name)
        access_val = f_index(i, f_s, f_num_derivs(_t))
        value = f_mul([f_component(f_transform, [f_r, f_s]), f_component(f_derivatives + _p, access_val)])
        lines += [f_iadd(name, value)]

    loop_vars = [(f_r, 0, f_num_derivs(_g)), (f_s, 0, f_num_derivs(_t))]
    code += f_loop(lines, loop_vars)
    return code



# -*- coding: utf-8 -*-

from ffc.uflacs.backends.ufc.generator import ufc_generator, ufc_integral_types

class ufc_integral(ufc_generator):
    def __init__(self, integral_type):
        assert integral_type in ufc_integral_types
        ufc_generator.__init__(self, integral_type + "_integral")

    def enabled_coefficients(self, L, ir):
        enabled_coefficients = ir["enabled_coefficients"]
        initializer_list = ", ".join("true" if enabled else "false"
                                     for enabled in enabled_coefficients)
        code = L.StatementList([
            # Cheating a bit with verbatim:
            L.VerbatimStatement("static const std::vector<bool> enabled({%s});" % initializer_list),
            L.Return(L.Symbol("enabled")),
            ])
        return code

    def tabulate_tensor(self, L, ir):
        # FIXME: This is where the current ffc backend code generation should be injected
        tt = ir["tabulate_tensor"]
        code = "code generated from %s" % tt
        return code

    def tabulate_tensor_comment(self, L, ir):
        # FIXME: Copy from ffc.codegeneration._generate_tabulate_tensor_comment
        return ""

class ufc_cell_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "cell")

class ufc_exterior_facet_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "exterior_facet")

class ufc_interior_facet_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "interior_facet")

class ufc_vertex_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "vertex")

class ufc_custom_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "custom")

    def num_cells(self, L, ir):
        value = ir["num_cells"]
        return L.Return(L.LiteralInt(value))

class ufc_interface_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "interface")

class ufc_overlap_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "overlap")

class ufc_cutcell_integral(ufc_integral):
    def __init__(self):
        ufc_integral.__init__(self, "cutcell")



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import pytest
import time

# FFC modules
from ffc.quadrature.reduce_operations import operation_count, expand_operations, reduce_operations
from ffc.quadrature.symbolics import *
from ffc.cpp import format, set_float_formatting
from ffc.parameters import FFC_PARAMETERS
set_float_formatting(FFC_PARAMETERS['precision'])


def testElasticity2D():
    elasticity = """(((Jinv_00*FE0_C0_D10_ip_j + Jinv_10*FE0_C0_D01_ip_j)*2*(Jinv_00*FE0_C0_D10_ip_k + Jinv_10*FE0_C0_D01_ip_k)*2 + ((Jinv_00*FE0_C1_D10_ip_j + Jinv_10*FE0_C1_D01_ip_j) + (Jinv_01*FE0_C0_D10_ip_j + Jinv_11*FE0_C0_D01_ip_j))*((Jinv_00*FE0_C1_D10_ip_k + Jinv_10*FE0_C1_D01_ip_k) + (Jinv_01*FE0_C0_D10_ip_k + Jinv_11*FE0_C0_D01_ip_k))) + ((Jinv_01*FE0_C1_D10_ip_j + Jinv_11*FE0_C1_D01_ip_j)*2*(Jinv_01*FE0_C1_D10_ip_k + Jinv_11*FE0_C1_D01_ip_k)*2 + ((Jinv_01*FE0_C0_D10_ip_j + Jinv_11*FE0_C0_D01_ip_j) + (Jinv_00*FE0_C1_D10_ip_j + Jinv_10*FE0_C1_D01_ip_j))*((Jinv_01*FE0_C0_D10_ip_k + Jinv_11*FE0_C0_D01_ip_k) + (Jinv_00*FE0_C1_D10_ip_k + Jinv_10*FE0_C1_D01_ip_k))))*0.25*W4_ip*det"""

    expr = Product([
        Sum([
            Sum([
                Product([
                    Sum([
                        Product([Symbol("Jinv_00", GEO), Symbol("FE0_C0_D10_ip_j", BASIS)]),
                        Product([Symbol("Jinv_10", GEO), Symbol("FE0_C0_D01_ip_j", BASIS)])
                    ]),
                    FloatValue(2),
                    Sum([
                        Product([Symbol("Jinv_00", GEO), Symbol("FE0_C0_D10_ip_k", BASIS)]),
                        Product([Symbol("Jinv_10", GEO), Symbol("FE0_C0_D01_ip_k", BASIS)])
                    ]),
                    FloatValue(2)
                ]),
                Product([
                        Sum([
                            Sum([
                                Product([Symbol("Jinv_00", GEO), Symbol("FE0_C1_D10_ip_j", BASIS)]),
                                Product([Symbol("Jinv_10", GEO), Symbol("FE0_C1_D01_ip_j", BASIS)])
                                ]),
                            Sum([
                                Product([Symbol("Jinv_01", GEO), Symbol("FE0_C0_D10_ip_j", BASIS)]),
                                Product([Symbol("Jinv_11", GEO), Symbol("FE0_C0_D01_ip_j", BASIS)])
                                ])
                            ]),
                Sum([
                    Sum([
                        Product([Symbol("Jinv_00", GEO), Symbol("FE0_C1_D10_ip_k", BASIS)]),
                        Product([Symbol("Jinv_10", GEO), Symbol("FE0_C1_D01_ip_k", BASIS)])
                    ]),
                    Sum([
                        Product([Symbol("Jinv_01", GEO), Symbol("FE0_C0_D10_ip_k", BASIS)]),
                        Product([Symbol("Jinv_11", GEO), Symbol("FE0_C0_D01_ip_k", BASIS)])
                    ])
                ])
                ])
            ]),
            Sum([
                Product([
                    Sum([
                        Product([Symbol("Jinv_01", GEO), Symbol("FE0_C1_D10_ip_j", BASIS)]),
                        Product([Symbol("Jinv_11", GEO), Symbol("FE0_C1_D01_ip_j", BASIS)])
                    ]),
                    FloatValue(2),
                    Sum([
                        Product([Symbol("Jinv_01", GEO), Symbol("FE0_C1_D10_ip_k", BASIS)]),
                        Product([Symbol("Jinv_11", GEO), Symbol("FE0_C1_D01_ip_k", BASIS)])
                    ]),
                    FloatValue(2)
                ]),
                Product([
                    Sum([
                        Sum([
                            Product([Symbol("Jinv_01", GEO), Symbol("FE0_C0_D10_ip_j", BASIS)]),
                            Product([Symbol("Jinv_11", GEO), Symbol("FE0_C0_D01_ip_j", BASIS)])
                        ]),
                        Sum([
                            Product([Symbol("Jinv_00", GEO), Symbol("FE0_C1_D10_ip_j", BASIS)]),
                            Product([Symbol("Jinv_10", GEO), Symbol("FE0_C1_D01_ip_j", BASIS)])
                        ])
                    ]),
                    Sum([
                        Sum([
                            Product([Symbol("Jinv_01", GEO), Symbol("FE0_C0_D10_ip_k", BASIS)]),
                            Product([Symbol("Jinv_11", GEO), Symbol("FE0_C0_D01_ip_k", BASIS)])
                        ]),
                        Sum([
                            Product([Symbol("Jinv_00", GEO), Symbol("FE0_C1_D10_ip_k", BASIS)]),
                            Product([Symbol("Jinv_10", GEO), Symbol("FE0_C1_D01_ip_k", BASIS)])
                        ])
                    ])
                ])
            ])
        ]),
        FloatValue(0.25),
        Symbol("W4_ip", IP),
        Symbol("det", GEO)
    ])

    expr_exp = expr.expand()
    elasticity_exp = expand_operations(elasticity, format)
    expr_red = expr_exp.reduce_ops()
    elasticity_red = reduce_operations(elasticity, format)
    elasticity_exp_ops = operation_count(elasticity_exp, format)
    elasticity_red_ops = operation_count(elasticity_red, format)

    Jinv_00, Jinv_01, Jinv_10, Jinv_11, W4_ip, det = (1.1, 1.5, -4.3, 1.7, 11, 52.3)
    FE0_C0_D01_ip_j, FE0_C0_D10_ip_j, FE0_C0_D01_ip_k, FE0_C0_D10_ip_k = (1.12, 5.7, -9.3, 7.4)
    FE0_C1_D01_ip_j, FE0_C1_D10_ip_j, FE0_C1_D01_ip_k, FE0_C1_D10_ip_k = (3.12, -8.1, -45.3, 17.5)

    assert round(eval(str(expr)) - eval(str(expr_exp)), 8) == 0.0
    assert round(eval(str(expr)) - eval(str(expr_red)), 8) == 0.0
    assert round(eval(str(expr)) - eval(str(elasticity)), 8) == 0.0
    assert round(eval(str(expr)) - eval(str(elasticity_exp)), 8) == 0.0
    assert round(eval(str(expr)) - eval(str(elasticity_red)), 8) == 0.0
    assert expr.ops() == 52
    assert elasticity_exp_ops == 159
    assert expr_exp.ops() == 159
    assert elasticity_red_ops == 71
    assert expr_red.ops() == 71



# -*- coding: utf-8 -*-
"""
This module contains functionality for working with automated
goal-oriented error control. In particular it offers the following
function:

  compile_with_error_control  - Compile forms and generate error control forms

"""
from .errorcontrol import compile_with_error_control
